# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import dataclasses
from collections import UserList
from typing import TYPE_CHECKING

from twisted.internet import defer
from twisted.python import log

from buildbot.data import resultspec
from buildbot.process.properties import renderer
from buildbot.process.results import RETRY
from buildbot.util import flatten

if TYPE_CHECKING:
    from buildbot.db.buildrequests import BuildRequestModel


@defer.inlineCallbacks
def getPreviousBuild(master, build):
    # naive n-1 algorithm. Still need to define what we should skip
    # SKIP builds? forced builds? rebuilds?
    # don't hesitate to contribute improvements to that algorithm
    n = build['number'] - 1
    while n >= 0:
        prev = yield master.data.get(("builders", build['builderid'], "builds", n))

        if prev and prev['results'] != RETRY:
            return prev
        n -= 1
    return None


@defer.inlineCallbacks
def getDetailsForBuildset(
    master,
    bsid,
    want_properties=False,
    want_steps=False,
    want_previous_build=False,
    want_logs=False,
    add_logs=None,
    want_logs_content=False,
):
    # Here we will do a bunch of data api calls on behalf of the reporters
    # We do try to make *some* calls in parallel with the help of gatherResults, but don't commit
    # to much in that. The idea is to do parallelism while keeping the code readable
    # and maintainable.

    # first, just get the buildset and all build requests for our buildset id
    dl = [
        master.data.get(("buildsets", bsid)),
        master.data.get(
            ('buildrequests',), filters=[resultspec.Filter('buildsetid', 'eq', [bsid])]
        ),
    ]
    (buildset, breqs) = yield defer.gatherResults(dl, consumeErrors=True)
    # next, get the bdictlist for each build request
    dl = [master.data.get(("buildrequests", breq['buildrequestid'], 'builds')) for breq in breqs]

    builds = yield defer.gatherResults(dl, consumeErrors=True)
    builds = flatten(builds, types=(list, UserList))
    if builds:
        yield getDetailsForBuilds(
            master,
            buildset,
            builds,
            want_properties=want_properties,
            want_steps=want_steps,
            want_previous_build=want_previous_build,
            want_logs=want_logs,
            add_logs=add_logs,
            want_logs_content=want_logs_content,
        )

    return {"buildset": buildset, "builds": builds}


@defer.inlineCallbacks
def getDetailsForBuild(
    master,
    build,
    want_properties=False,
    want_steps=False,
    want_previous_build=False,
    want_logs=False,
    add_logs=None,
    want_logs_content=False,
):
    buildrequest = yield master.data.get(("buildrequests", build['buildrequestid']))
    buildset = yield master.data.get(("buildsets", buildrequest['buildsetid']))
    build['buildrequest'] = buildrequest
    build['buildset'] = buildset

    parentbuild = None
    parentbuilder = None
    if buildset['parent_buildid']:
        parentbuild = yield master.data.get(("builds", buildset['parent_buildid']))
        parentbuilder = yield master.data.get(("builders", parentbuild['builderid']))
    build['parentbuild'] = parentbuild
    build['parentbuilder'] = parentbuilder

    ret = yield getDetailsForBuilds(
        master,
        buildset,
        [build],
        want_properties=want_properties,
        want_steps=want_steps,
        want_previous_build=want_previous_build,
        want_logs=want_logs,
        add_logs=add_logs,
        want_logs_content=want_logs_content,
    )
    return ret


@defer.inlineCallbacks
def get_details_for_buildrequest(master, buildrequest: BuildRequestModel, build):
    buildset = yield master.data.get(("buildsets", buildrequest.buildsetid))
    builder = yield master.data.get(("builders", buildrequest.builderid))

    build['buildrequest'] = dataclasses.asdict(buildrequest)
    build['buildset'] = buildset
    build['builderid'] = buildrequest.builderid
    build['builder'] = builder
    build['url'] = getURLForBuildrequest(master, buildrequest.buildrequestid)
    build['results'] = None
    build['complete'] = False


def should_attach_log(logs_config, log):
    if isinstance(logs_config, bool):
        return logs_config

    if log['name'] in logs_config:
        return True

    long_name = f"{log['stepname']}.{log['name']}"
    if long_name in logs_config:
        return True

    return False


@defer.inlineCallbacks
def getDetailsForBuilds(
    master,
    buildset,
    builds,
    want_properties=False,
    want_steps=False,
    want_previous_build=False,
    want_logs=False,
    add_logs=None,
    want_logs_content=False,
):
    builderids = {build['builderid'] for build in builds}

    builders = yield defer.gatherResults(
        [master.data.get(("builders", _id)) for _id in builderids], consumeErrors=True
    )

    buildersbyid = {builder['builderid']: builder for builder in builders}

    if want_properties:
        buildproperties = yield defer.gatherResults(
            [master.data.get(("builds", build['buildid'], 'properties')) for build in builds],
            consumeErrors=True,
        )
    else:  # we still need a list for the big zip
        buildproperties = list(range(len(builds)))

    if want_previous_build:
        prev_builds = yield defer.gatherResults(
            [getPreviousBuild(master, build) for build in builds], consumeErrors=True
        )
    else:  # we still need a list for the big zip
        prev_builds = list(range(len(builds)))

    if add_logs is not None:
        logs_config = add_logs
    elif want_logs_content is not None:
        logs_config = want_logs_content
    else:
        logs_config = False

    if logs_config is not False:
        want_logs = True
    if want_logs:
        want_steps = True

    if want_steps:  # pylint: disable=too-many-nested-blocks
        buildsteps = yield defer.gatherResults(
            [master.data.get(("builds", build['buildid'], 'steps')) for build in builds],
            consumeErrors=True,
        )
        if want_logs:
            for build, build_steps in zip(builds, buildsteps):
                for s in build_steps:
                    logs = yield master.data.get(("steps", s['stepid'], 'logs'))
                    s['logs'] = list(logs)
                    for l in s['logs']:
                        l['stepname'] = s['name']
                        l['url'] = get_url_for_log(
                            master, build['builderid'], build['number'], s['number'], l['slug']
                        )
                        l['url_raw'] = get_url_for_log_raw(master, l['logid'], 'raw')
                        l['url_raw_inline'] = get_url_for_log_raw(master, l['logid'], 'raw_inline')
                        if should_attach_log(logs_config, l):
                            l['content'] = yield master.data.get(("logs", l['logid'], 'contents'))

    else:  # we still need a list for the big zip
        buildsteps = list(range(len(builds)))

    # a big zip to connect everything together
    for build, properties, steps, prev in zip(builds, buildproperties, buildsteps, prev_builds):
        build['builder'] = buildersbyid[build['builderid']]
        build['buildset'] = buildset
        build['url'] = getURLForBuild(master, build['builderid'], build['number'])

        if want_properties:
            build['properties'] = properties

        if want_steps:
            build['steps'] = list(steps)

        if want_previous_build:
            build['prev_build'] = prev


# perhaps we need data api for users with sourcestamps/:id/users
@defer.inlineCallbacks
def getResponsibleUsersForSourceStamp(master, sourcestampid):
    changesd = master.data.get(("sourcestamps", sourcestampid, "changes"))
    sourcestampd = master.data.get(("sourcestamps", sourcestampid))
    changes, sourcestamp = yield defer.gatherResults([changesd, sourcestampd], consumeErrors=True)
    blamelist = set()
    # normally, we get only one, but just assume there might be several
    for c in changes:
        blamelist.add(c['author'])
    # Add patch author to blamelist
    if 'patch' in sourcestamp and sourcestamp['patch'] is not None:
        blamelist.add(sourcestamp['patch']['author'])
    blamelist = list(blamelist)
    blamelist.sort()
    return blamelist


# perhaps we need data api for users with builds/:id/users
@defer.inlineCallbacks
def getResponsibleUsersForBuild(master, buildid):
    dl = [
        master.data.get(("builds", buildid, "changes")),
        master.data.get(("builds", buildid, 'properties')),
    ]
    changes, properties = yield defer.gatherResults(dl, consumeErrors=True)
    blamelist = set()

    # add users from changes
    for c in changes:
        blamelist.add(c['author'])

    # add owner from properties
    if 'owner' in properties:
        owner = properties['owner'][0]
        if isinstance(owner, str):
            blamelist.add(owner)
        else:
            blamelist.update(owner)
            log.msg(f"Warning: owner property is a list for buildid {buildid}. ")
            log.msg(f"Please report a bug: changes: {changes}. properties: {properties}")

    # add owner from properties
    if 'owners' in properties:
        blamelist.update(properties['owners'][0])

    blamelist = list(blamelist)
    blamelist.sort()
    return blamelist


# perhaps we need data api for users with buildsets/:id/users
@defer.inlineCallbacks
def get_responsible_users_for_buildset(master, buildsetid):
    props = yield master.data.get(("buildsets", buildsetid, "properties"))

    # TODO: This currently does not track what changes were in the buildset. getChangesForBuild()
    # would walk the change graph until it finds last successful build and uses the authors of
    # the changes as blame list. Probably this needs to be done here too
    owner = props.get("owner", None)
    if owner:
        return [owner[0]]
    return []


def getURLForBuild(master, builderid, build_number):
    prefix = master.config.buildbotURL
    return prefix + f"#/builders/{builderid}/builds/{build_number}"


def getURLForBuildrequest(master, buildrequestid):
    prefix = master.config.buildbotURL
    return f"{prefix}#/buildrequests/{buildrequestid}"


def get_url_for_log(master, builderid, build_number, step_number, log_slug):
    prefix = master.config.buildbotURL
    return (
        f"{prefix}#/builders/{builderid}/builds/{build_number}/"
        + f"steps/{step_number}/logs/{log_slug}"
    )


def get_url_for_log_raw(master, logid, suffix):
    prefix = master.config.buildbotURL
    return f"{prefix}api/v2/logs/{logid}/{suffix}"


@renderer
def URLForBuild(props):
    build = props.getBuild()
    return build.getUrl()


def merge_reports_prop(reports, prop):
    result = None
    for report in reports:
        if prop in report and report[prop] is not None:
            if result is None:
                result = report[prop]
            else:
                result += report[prop]

    return result


def merge_reports_prop_take_first(reports, prop):
    for report in reports:
        if prop in report and report[prop] is not None:
            return report[prop]

    return None



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from buildbot.db import base
from buildbot.util.sautils import hash_columns

if TYPE_CHECKING:
    import sqlalchemy as sa
    from twisted.internet import defer

    from buildbot.data.resultspec import ResultSpec


@dataclass
class CodebaseModel:
    id: int
    projectid: int
    name: str
    slug: str


class CodebasesConnectorComponent(base.DBConnectorComponent):
    def find_codebase_id(
        self, *, projectid: int, name: str, auto_create: bool = True
    ) -> defer.Deferred[int | None]:
        tbl = self.db.model.codebases
        name_hash = hash_columns(name)
        return self.findSomethingId(
            tbl=tbl,
            whereclause=((tbl.c.projectid == projectid) & (tbl.c.name_hash == name_hash)),
            insert_values={
                'name': name,
                'name_hash': name_hash,
                'slug': name,
                'projectid': projectid,
            },
            autoCreate=auto_create,
        )

    def get_codebase(self, codebaseid: int) -> defer.Deferred[CodebaseModel | None]:
        def thd(conn: sa.engine.Connection) -> CodebaseModel | None:
            q = self.db.model.codebases.select().where(
                self.db.model.codebases.c.id == codebaseid,
            )
            res = conn.execute(q)
            row = res.fetchone()

            rv = None
            if row:
                rv = self._model_from_row(row)
            res.close()
            return rv

        return self.db.pool.do(thd)

    def get_codebases(
        self, projectid: int | None = None, result_spec: ResultSpec | None = None
    ) -> defer.Deferred[list[CodebaseModel]]:
        def thd(conn: sa.engine.Connection) -> list[CodebaseModel]:
            tbl = self.db.model.codebases
            q = tbl.select()
            if projectid is not None:
                q = q.where(tbl.c.projectid == projectid)
            if result_spec is not None:
                return result_spec.thd_execute(conn, q, lambda row: self._model_from_row(row))
            res = conn.execute(q)
            return [self._model_from_row(row) for row in res.fetchall()]

        return self.db.pool.do(thd)

    def update_codebase_info(
        self,
        *,
        codebaseid: int,
        projectid: int,
        slug: str,
    ) -> defer.Deferred[None]:
        def thd(conn: sa.engine.Connection) -> None:
            q = self.db.model.codebases.update().where(self.db.model.codebases.c.id == codebaseid)
            conn.execute(
                q.values(
                    projectid=projectid,
                    slug=slug,
                )
            ).close()

        return self.db.pool.do_with_transaction(thd)

    def _model_from_row(self, row: sa.engine.Row) -> CodebaseModel:
        return CodebaseModel(
            id=row.id,
            projectid=row.projectid,
            name=row.name,
            slug=row.slug,
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.application import service
from twisted.internet import defer

from buildbot.util import service as util_service


class UserManagerManager(util_service.ReconfigurableServiceMixin, service.MultiService):
    # this class manages a fleet of user managers; hence the name..

    def __init__(self, master):
        super().__init__()
        self.setName('user_manager_manager')
        self.master = master

    @defer.inlineCallbacks
    def reconfigServiceWithBuildbotConfig(self, new_config):
        # this is easy - kick out all of the old managers, and add the
        # new ones.

        # pylint: disable=cell-var-from-loop
        for mgr in list(self):
            yield mgr.disownServiceParent()

        for mgr in new_config.user_managers:
            yield mgr.setServiceParent(self)

        # reconfig any newly-added change sources, as well as existing
        yield super().reconfigServiceWithBuildbotConfig(new_config)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import TYPE_CHECKING

from twisted.internet import defer
from twisted.python import log

from buildbot import locks
from buildbot import util
from buildbot.db.buildrequests import AlreadyClaimedError
from buildbot.process import metrics
from buildbot.process.builder import Builder
from buildbot.process.buildrequestdistributor import BuildRequestDistributor
from buildbot.process.results import CANCELLED
from buildbot.process.results import RETRY
from buildbot.process.workerforbuilder import States
from buildbot.util import debounce
from buildbot.util import service
from buildbot.util.render_description import render_description
from buildbot.util.twisted import async_to_deferred
from buildbot.worker.latent import AbstractLatentWorker

if TYPE_CHECKING:
    from buildbot.worker import AbstractWorker


class LockRetrieverMixin:
    @defer.inlineCallbacks
    def getLockByID(self, lockid, config_version):
        """Convert a Lock identifier into an actual Lock instance.
        @lockid: a locks.MasterLock or locks.WorkerLock instance
        @config_version: The version of the config from which the list of locks has been
            acquired by the downstream user.
        @return: a locks.RealMasterLock or locks.RealWorkerLock instance

        The real locks are tracked using lock ID and config_version. The latter is used as a
        way to track most recent properties of real locks.

        This approach is needed because there's no central registry of lock access instances
        that are used within a Buildbot master.cfg (like there is for e.g c['builders']). All
        lock accesses bring all lock information with themselves as the lockid member.
        Therefore, the reconfig process is relatively complicated, because we don't know
        whether a specific access instance encodes lock information before reconfig or after.
        Taking into account config_version allows us to know when properties of a lock should
        be updated.

        Note that the user may create multiple lock ids with different maxCount values. It's
        unspecified which maxCount value the real lock will have.
        """
        assert isinstance(config_version, int)
        lock = yield lockid.lockClass.getService(self, lockid.name)

        if config_version > lock.config_version:
            lock.updateFromLockId(lockid, config_version)
        return lock

    def getLockFromLockAccess(self, access, config_version):
        # Convert a lock-access object into an actual Lock instance.
        if not isinstance(access, locks.LockAccess):
            # Buildbot 0.7.7 compatibility: user did not specify access
            access = access.defaultAccess()
        return self.getLockByID(access.lockid, config_version)

    @defer.inlineCallbacks
    def getLockFromLockAccesses(self, accesses, config_version):
        # converts locks to their real forms
        locks = yield defer.gatherResults(
            [self.getLockFromLockAccess(access, config_version) for access in accesses],
            consumeErrors=True,
        )
        return zip(locks, accesses)


class BotMaster(service.ReconfigurableServiceMixin, service.AsyncMultiService, LockRetrieverMixin):
    """This is the master-side service which manages remote buildbot workers.
    It provides them with Workers, and distributes build requests to
    them."""

    debug = 0
    name: str | None = "botmaster"  # type: ignore[assignment]

    def __init__(self) -> None:
        super().__init__()

        self.builders: dict[str, Builder] = {}
        self.builderNames: list[str] = []
        # builders maps Builder names to instances of bb.p.builder.Builder,
        # which is the master-side object that defines and controls a build.

        # Unused?
        self.watchers: dict[object, object] = {}

        self.shuttingDown = False

        # subscription to new build requests
        self.buildrequest_consumer_new = None
        self.buildrequest_consumer_unclaimed = None
        self.buildrequest_consumer_cancel = None

        # a distributor for incoming build requests; see below
        self.brd = BuildRequestDistributor(self)
        self.brd.setServiceParent(self)
        self._pending_builderids: set[str] = set()

        # Dictionary of build request ID to False or cancellation reason string in case cancellation
        # has been requested.
        self._starting_brid_to_cancel: dict[int, bool | str] = {}

    @defer.inlineCallbacks
    def cleanShutdown(self, quickMode=False, stopReactor=True):
        """Shut down the entire process, once all currently-running builds are
        complete.
        quickMode will mark all builds as retry (except the ones that were triggered)
        """
        if self.shuttingDown:
            return
        log.msg("Initiating clean shutdown")
        self.shuttingDown = True
        # first, stop the distributor; this will finish any ongoing scheduling
        # operations before firing
        if quickMode:
            # if quick mode, builds will be cancelled, so stop scheduling altogether
            yield self.brd.disownServiceParent()
        else:
            # if not quick, still schedule waited child builds
            # other parent will never finish
            self.brd.distribute_only_waited_childs = True

        # Double check that we're still supposed to be shutting down
        # The shutdown may have been cancelled!
        while self.shuttingDown:
            if quickMode:
                for builder in self.builders.values():
                    # As we stop the builds, builder.building might change during loop
                    # so we need to copy the list
                    for build in list(builder.building):
                        # if build is waited for then this is a sub-build, so
                        # no need to retry it
                        if sum(br.waitedFor for br in build.requests):
                            results = CANCELLED
                        else:
                            results = RETRY
                        is_building = (
                            build.workerforbuilder is not None
                            and build.workerforbuilder.state == States.BUILDING
                        )

                        # Master should not wait build.stopBuild for ages to complete if worker
                        # does not send any message about shutting the builds down quick enough.
                        # Just kill the connection with the worker
                        def lose_connection(b):
                            if b.workerforbuilder.worker.conn is not None:
                                b.workerforbuilder.worker.conn.loseConnection()

                        sheduled_call = self.master.reactor.callLater(5, lose_connection, build)

                        def cancel_lose_connection(_, call):
                            if call.active():
                                call.cancel()

                        d = build.stopBuild("Master Shutdown", results)
                        d.addBoth(cancel_lose_connection, sheduled_call)

                        if not is_building:
                            # if it is not building, then it must be a latent worker
                            # which is substantiating. Cancel it.
                            if build.workerforbuilder is not None and isinstance(
                                build.workerforbuilder.worker,
                                AbstractLatentWorker,
                            ):
                                build.workerforbuilder.worker.insubstantiate()
            # then wait for all builds to finish
            dl = []
            for builder in self.builders.values():
                for build in builder.building:
                    # build may be waiting for ping to worker to succeed which
                    # may never happen if the connection to worker was broken
                    # without TCP connection being severed
                    build.workerforbuilder.abortPingIfAny()

                    dl.append(build.waitUntilFinished())
            if not dl:
                log.msg("No running jobs, starting shutdown immediately")
            else:
                log.msg(f"Waiting for {len(dl)} build(s) to finish")
                yield defer.DeferredList(dl, consumeErrors=True)

            # Check that there really aren't any running builds
            n = 0
            for builder in self.builders.values():
                if builder.building:
                    num_builds = len(builder.building)
                    log.msg(f"Builder {builder} has {num_builds} builds running")
                    n += num_builds
            if n > 0:
                log.msg(f"Not shutting down, there are {n} builds running")
                log.msg("Trying shutdown sequence again")
                yield util.asyncSleep(1)
            else:
                break

        # shutdown was cancelled
        if not self.shuttingDown:
            if quickMode:
                yield self.brd.setServiceParent(self)
            else:
                self.brd.distribute_only_waited_childs = False

            return

        if stopReactor:
            log.msg("Stopping reactor")
            self.master.reactor.stop()

    def cancelCleanShutdown(self):
        """Cancel a clean shutdown that is already in progress, if any"""
        if not self.shuttingDown:
            return
        log.msg("Cancelling clean shutdown")
        self.shuttingDown = False

    @metrics.countMethod('BotMaster.workerLost()')
    def workerLost(self, bot: AbstractWorker):
        metrics.MetricCountEvent.log("BotMaster.attached_workers", -1)
        for b in self.builders.values():
            if b.config is not None and bot.workername in b.config.workernames:
                b.detached(bot)

    @metrics.countMethod('BotMaster.getBuildersForWorker()')
    def getBuildersForWorker(self, workername: str):
        return [
            b
            for b in self.builders.values()
            if b.config is not None and workername in b.config.workernames
        ]

    def getBuildernames(self):
        return self.builderNames

    def getBuilders(self):
        return list(self.builders.values())

    def _buildrequest_added(self, key, msg):
        self._pending_builderids.add(msg['builderid'])
        self._flush_pending_builders()

    # flush pending builders needs to be debounced, as per design the
    # buildrequests events will arrive in burst.
    # We debounce them to let the brd manage them as a whole
    # without having to debounce the brd itself
    @debounce.method(wait=0.1, until_idle=True)
    def _flush_pending_builders(self):
        if not self._pending_builderids:
            return
        buildernames = []
        for builderid in self._pending_builderids:
            builder = self.getBuilderById(builderid)
            if builder:
                buildernames.append(builder.name)
        self._pending_builderids.clear()
        self.brd.maybeStartBuildsOn(buildernames)

    def getBuilderById(self, builderid):
        return self._builders_byid.get(builderid)

    @defer.inlineCallbacks
    def startService(self):
        # consume both 'new' and 'unclaimed' build requests events
        startConsuming = self.master.mq.startConsuming
        self.buildrequest_consumer_new = yield startConsuming(
            self._buildrequest_added, ('buildrequests', None, "new")
        )
        self.buildrequest_consumer_unclaimed = yield startConsuming(
            self._buildrequest_added, ('buildrequests', None, 'unclaimed')
        )
        self.buildrequest_consumer_cancel = yield startConsuming(
            self._buildrequest_canceled, ('control', 'buildrequests', None, 'cancel')
        )
        yield super().startService()

    @defer.inlineCallbacks
    def _buildrequest_canceled(self, key, msg):
        brid = int(key[2])
        reason = msg.get('reason', 'no reason')

        # first, try to claim the request; if this fails, then it's too late to
        # cancel the build anyway
        try:
            b = yield self.master.db.buildrequests.claimBuildRequests(brids=[brid])
        except AlreadyClaimedError:
            self.maybe_cancel_in_progress_buildrequest(brid, reason)

            # In case the build request has been claimed on this master, the call to
            # maybe_cancel_in_progress_buildrequest above will ensure that they are either visible
            # to the data API call below, or canceled.
            builds = yield self.master.data.get(("buildrequests", brid, "builds"))

            # Any other master will observe the buildrequest cancel messages and will try to
            # cancel the buildrequest or builds internally.
            #
            # TODO: do not try to cancel builds that run on another master. Note that duplicate
            # cancels do not have any downside.
            for b in builds:
                self.master.mq.produce(
                    ("control", "builds", str(b['buildid']), "stop"), {'reason': reason}
                )
            return

        # then complete it with 'CANCELLED'; this is the closest we can get to
        # cancelling a request without running into trouble with dangling
        # references.
        yield self.master.data.updates.completeBuildRequests([brid], CANCELLED)
        brdict = yield self.master.db.buildrequests.getBuildRequest(brid)
        self.master.mq.produce(('buildrequests', str(brid), 'cancel'), brdict)

    @defer.inlineCallbacks
    def reconfigServiceWithBuildbotConfig(self, new_config):
        timer = metrics.Timer("BotMaster.reconfigServiceWithBuildbotConfig")
        timer.start()

        yield self.reconfigProjects(new_config)
        yield self.reconfig_codebases(new_config)
        yield self.reconfigServiceBuilders(new_config)

        # call up
        yield super().reconfigServiceWithBuildbotConfig(new_config)

        # try to start a build for every builder; this is necessary at master
        # startup, and a good idea in any other case
        self.maybeStartBuildsForAllBuilders()

        timer.stop()

    @defer.inlineCallbacks
    def reconfigProjects(self, new_config):
        for project_config in new_config.projects:
            projectid = yield self.master.data.updates.find_project_id(project_config.name)
            yield self.master.data.updates.update_project_info(
                projectid,
                project_config.slug,
                project_config.description,
                project_config.description_format,
                render_description(project_config.description, project_config.description_format),
            )

    @async_to_deferred
    async def reconfig_codebases(self, new_config):
        for codebase_config in new_config.codebases:
            projectid = await self.master.data.updates.find_project_id(
                codebase_config.project, auto_create=False
            )
            if projectid is None:
                raise RuntimeError(
                    f'Could not find project {codebase_config.project} '
                    f'for codebase {codebase_config.name}'
                )

            codebaseid = await self.master.data.updates.find_codebase_id(
                projectid=projectid,
                name=codebase_config.name,
            )

            await self.master.data.updates.update_codebase_info(
                codebaseid=codebaseid,
                projectid=projectid,
                slug=codebase_config.slug,
            )

    @defer.inlineCallbacks
    def reconfigServiceBuilders(self, new_config):
        timer = metrics.Timer("BotMaster.reconfigServiceBuilders")
        timer.start()

        # arrange builders by name
        old_by_name = {b.name: b for b in list(self) if isinstance(b, Builder)}
        old_set = set(old_by_name)
        new_by_name = {bc.name: bc for bc in new_config.builders}
        new_set = set(new_by_name)

        # calculate new builders, by name, and removed builders
        removed_names, added_names = util.diffSets(old_set, new_set)

        if removed_names or added_names:
            log.msg(f"adding {len(added_names)} new builders, removing {len(removed_names)}")

            for n in removed_names:
                builder = old_by_name[n]

                del self.builders[n]
                builder.master = None
                builder.botmaster = None

                yield builder.disownServiceParent()

            for n in added_names:
                builder = Builder(n)
                self.builders[n] = builder

                builder.botmaster = self
                builder.master = self.master
                yield builder.setServiceParent(self)

        self.builderNames = list(self.builders)
        self._builders_byid = {}
        for builder in self.builders.values():
            self._builders_byid[(yield builder.getBuilderId())] = builder

        yield self.master.data.updates.updateBuilderList(
            self.master.masterid, [util.bytes2unicode(n) for n in self.builderNames]
        )

        metrics.MetricCountEvent.log("num_builders", len(self.builders), absolute=True)

        timer.stop()

    def stopService(self):
        if self.buildrequest_consumer_new:
            self.buildrequest_consumer_new.stopConsuming()
            self.buildrequest_consumer_new = None
        if self.buildrequest_consumer_unclaimed:
            self.buildrequest_consumer_unclaimed.stopConsuming()
            self.buildrequest_consumer_unclaimed = None
        if self.buildrequest_consumer_cancel:
            self.buildrequest_consumer_cancel.stopConsuming()
            self.buildrequest_consumer_cancel = None
        self._pending_builderids.clear()
        self._flush_pending_builders.stop()
        return super().stopService()

    # Used to track buildrequests that are in progress of being started on this master.
    def add_in_progress_buildrequest(self, brid):
        self._starting_brid_to_cancel[brid] = False

    def remove_in_progress_buildrequest(self, brid):
        return self._starting_brid_to_cancel.pop(brid, None)

    def maybe_cancel_in_progress_buildrequest(self, brid, reason):
        """
        Ensures that after this call any builds resulting from build request will be visible or
        cancelled.
        """
        if brid in self._starting_brid_to_cancel:
            self._starting_brid_to_cancel[brid] = reason

    def maybeStartBuildsForBuilder(self, buildername):
        """
        Call this when something suggests that a particular builder may now
        be available to start a build.

        @param buildername: the name of the builder
        """
        self.brd.maybeStartBuildsOn([buildername])

    def maybeStartBuildsForWorker(self, worker_name):
        """
        Call this when something suggests that a particular worker may now be
        available to start a build.

        @param worker_name: the name of the worker
        """
        builders = self.getBuildersForWorker(worker_name)
        self.brd.maybeStartBuildsOn([b.name for b in builders])

    def maybeStartBuildsForAllBuilders(self):
        """
        Call this when something suggests that this would be a good time to
        start some builds, but nothing more specific.
        """
        self.brd.maybeStartBuildsOn(self.builderNames)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import os
from unittest import mock

from twisted.internet import defer
from twisted.internet import reactor
from twisted.python import log
from twisted.python.filepath import FilePath

from buildbot import util
from buildbot.clients import tryclient
from buildbot.schedulers import trysched
from buildbot.test.util import www
from buildbot.test.util.integration import RunMasterBase


# wait for some asynchronous result
@defer.inlineCallbacks
def waitFor(fn):
    while True:
        res = yield fn()
        if res:
            return res
        yield util.asyncSleep(0.01)


class Schedulers(RunMasterBase, www.RequiresWwwMixin):
    def setUp(self):
        self.master = None
        self.sch = None

        def spawnProcess(pp, executable, args, environ):
            tmpfile = os.path.join(self.jobdir, 'tmp', 'testy')
            newfile = os.path.join(self.jobdir, 'new', 'testy')
            with open(tmpfile, "w", encoding='utf-8') as f:
                f.write(pp.job)
            os.rename(tmpfile, newfile)
            log.msg(f"wrote jobfile {newfile}")
            # get the scheduler to poll this directory now
            d = self.sch.watcher.poll()
            d.addErrback(log.err, 'while polling')

            @d.addCallback
            def finished(_):
                st = mock.Mock()
                st.value.signal = None
                st.value.exitCode = 0
                pp.processEnded(st)

        self.patch(reactor, 'spawnProcess', spawnProcess)

        self.sourcestamp = tryclient.SourceStamp(branch='br', revision='rr', patch=(0, '++--'))

        def getSourceStamp(vctype, treetop, branch=None, repository=None):
            return defer.succeed(self.sourcestamp)

        self.patch(tryclient, 'getSourceStamp', getSourceStamp)

        self.output = []

        # stub out printStatus, as it's timing-based and thus causes
        # occasional test failures.
        self.patch(tryclient.Try, 'printStatus', lambda _: None)

        def output(*msg):
            msg = ' '.join(map(str, msg))
            log.msg(f"output: {msg}")
            self.output.append(msg)

        self.patch(tryclient, 'output', output)

    def setupJobdir(self):
        jobdir = FilePath(self.mktemp())
        jobdir.createDirectory()
        self.jobdir = jobdir.path
        for sub in 'new', 'tmp', 'cur':
            jobdir.child(sub).createDirectory()
        return self.jobdir

    @defer.inlineCallbacks
    def setup_config(self, extra_config):
        c = {}
        from buildbot.config import BuilderConfig
        from buildbot.process import results
        from buildbot.process.buildstep import BuildStep
        from buildbot.process.factory import BuildFactory

        class MyBuildStep(BuildStep):
            def run(self):
                return results.SUCCESS

        c['change_source'] = []
        c['schedulers'] = []  # filled in above
        f1 = BuildFactory()
        f1.addStep(MyBuildStep(name='one'))
        f1.addStep(MyBuildStep(name='two'))
        c['builders'] = [
            BuilderConfig(name="a", workernames=["local1"], factory=f1),
        ]
        c['title'] = "test"
        c['titleURL'] = "test"
        c['buildbotURL'] = "http://localhost:8010/"
        c['mq'] = {'debug': True}
        # test wants to influence the config, but we still return a new config
        # each time
        c.update(extra_config)
        yield self.setup_master(c)

    @defer.inlineCallbacks
    def startMaster(self, sch):
        extra_config = {
            'schedulers': [sch],
        }
        self.sch = sch

        yield self.setup_config(extra_config)

        # wait until the scheduler is active
        yield waitFor(lambda: self.sch.active)

        # and, for Try_Userpass, until it's registered its port
        if isinstance(self.sch, trysched.Try_Userpass):

            def getSchedulerPort():
                if not self.sch.registrations:
                    return None
                self.serverPort = self.sch.registrations[0].getPort()
                log.msg(f"Scheduler registered at port {self.serverPort}")
                return True

            yield waitFor(getSchedulerPort)

    def runClient(self, config):
        self.clt = tryclient.Try(config)
        return self.clt.run_impl()

    @defer.inlineCallbacks
    def test_userpass_no_wait(self):
        yield self.startMaster(
            trysched.Try_Userpass(name='try', builderNames=['a'], port=0, userpass=[('u', b'p')])
        )
        yield self.runClient({
            'connect': 'pb',
            'master': f'127.0.0.1:{self.serverPort}',
            'username': 'u',
            'passwd': b'p',
        })
        self.assertEqual(
            self.output,
            [
                "using 'pb' connect method",
                'job created',
                'Delivering job; comment= None',
                'job has been delivered',
                'not waiting for builds to finish',
            ],
        )
        buildsets = yield self.master.db.buildsets.getBuildsets()
        self.assertEqual(len(buildsets), 1)

    @defer.inlineCallbacks
    def test_userpass_wait(self):
        yield self.startMaster(
            trysched.Try_Userpass(name='try', builderNames=['a'], port=0, userpass=[('u', b'p')])
        )
        yield self.runClient({
            'connect': 'pb',
            'master': f'127.0.0.1:{self.serverPort}',
            'username': 'u',
            'passwd': b'p',
            'wait': True,
        })
        self.assertEqual(
            self.output,
            [
                "using 'pb' connect method",
                'job created',
                'Delivering job; comment= None',
                'job has been delivered',
                'All Builds Complete',
                'a: success (build successful)',
            ],
        )
        buildsets = yield self.master.db.buildsets.getBuildsets()
        self.assertEqual(len(buildsets), 1)

    @defer.inlineCallbacks
    def test_userpass_wait_bytes(self):
        self.sourcestamp = tryclient.SourceStamp(branch=b'br', revision=b'rr', patch=(0, b'++--'))

        yield self.startMaster(
            trysched.Try_Userpass(name='try', builderNames=['a'], port=0, userpass=[('u', b'p')])
        )
        yield self.runClient({
            'connect': 'pb',
            'master': f'127.0.0.1:{self.serverPort}',
            'username': 'u',
            'passwd': b'p',
            'wait': True,
        })
        self.assertEqual(
            self.output,
            [
                "using 'pb' connect method",
                'job created',
                'Delivering job; comment= None',
                'job has been delivered',
                'All Builds Complete',
                'a: success (build successful)',
            ],
        )
        buildsets = yield self.master.db.buildsets.getBuildsets()
        self.assertEqual(len(buildsets), 1)

    @defer.inlineCallbacks
    def test_userpass_wait_dryrun(self):
        yield self.startMaster(
            trysched.Try_Userpass(name='try', builderNames=['a'], port=0, userpass=[('u', b'p')])
        )
        yield self.runClient({
            'connect': 'pb',
            'master': f'127.0.0.1:{self.serverPort}',
            'username': 'u',
            'passwd': b'p',
            'wait': True,
            'dryrun': True,
        })
        self.assertEqual(
            self.output,
            [
                "using 'pb' connect method",
                'job created',
                'Job:\n'
                '\tRepository: \n'
                '\tProject: \n'
                '\tBranch: br\n'
                '\tRevision: rr\n'
                '\tBuilders: None\n'
                '++--',
                'job has been delivered',
                'All Builds Complete',
            ],
        )
        buildsets = yield self.master.db.buildsets.getBuildsets()
        self.assertEqual(len(buildsets), 0)

    @defer.inlineCallbacks
    def test_userpass_list_builders(self):
        yield self.startMaster(
            trysched.Try_Userpass(name='try', builderNames=['a'], port=0, userpass=[('u', b'p')])
        )
        yield self.runClient({
            'connect': 'pb',
            'get-builder-names': True,
            'master': f'127.0.0.1:{self.serverPort}',
            'username': 'u',
            'passwd': b'p',
        })
        self.assertEqual(
            self.output,
            [
                "using 'pb' connect method",
                'The following builders are available for the try scheduler: ',
                'a',
            ],
        )
        buildsets = yield self.master.db.buildsets.getBuildsets()
        self.assertEqual(len(buildsets), 0)

    @defer.inlineCallbacks
    def test_jobdir_no_wait(self):
        jobdir = self.setupJobdir()
        yield self.startMaster(trysched.Try_Jobdir(name='try', builderNames=['a'], jobdir=jobdir))
        yield self.runClient({
            'connect': 'ssh',
            'master': '127.0.0.1',
            'username': 'u',
            'passwd': b'p',
            'builders': 'a',  # appears to be required for ssh
        })
        self.assertEqual(
            self.output,
            [
                "using 'ssh' connect method",
                'job created',
                'job has been delivered',
                'not waiting for builds to finish',
            ],
        )
        buildsets = yield self.master.db.buildsets.getBuildsets()
        self.assertEqual(len(buildsets), 1)

    @defer.inlineCallbacks
    def test_jobdir_wait(self):
        jobdir = self.setupJobdir()
        yield self.startMaster(trysched.Try_Jobdir(name='try', builderNames=['a'], jobdir=jobdir))
        yield self.runClient({
            'connect': 'ssh',
            'wait': True,
            'host': '127.0.0.1',
            'username': 'u',
            'passwd': b'p',
            'builders': 'a',  # appears to be required for ssh
        })
        self.assertEqual(
            self.output,
            [
                "using 'ssh' connect method",
                'job created',
                'job has been delivered',
                'waiting for builds with ssh is not supported',
            ],
        )
        buildsets = yield self.master.db.buildsets.getBuildsets()
        self.assertEqual(len(buildsets), 1)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

# This module is for backward compatibility of importlib.


def entry_points_get(entry_points, group):
    """Since Python 3.12 dictionary access is removed and replaced by new interface.
    see: https://github.com/python/cpython/issues/97781
    """
    if hasattr(entry_points, "select"):
        return entry_points.select(group=group)
    else:
        if isinstance(entry_points, list):
            filtered_entry_points = []
            for ep in entry_points:
                if ep.group == group:
                    filtered_entry_points.append(ep)
            return filtered_entry_points
        else:
            return entry_points.get(group, [])



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import TYPE_CHECKING

from twisted.internet import defer

from buildbot.data import base
from buildbot.data import types

if TYPE_CHECKING:
    from typing import Any
    from typing import AsyncGenerator

    from buildbot.db.logs import LogModel


class LogChunkEndpointBase(base.BuildNestingMixin, base.Endpoint):
    @staticmethod
    async def get_raw_log_lines(log_lines_generator: AsyncGenerator[str, None]) -> str:
        parts = []
        async for chunk in log_lines_generator:
            parts.append(chunk)
        return ''.join(parts)

    async def get_log_lines(
        self,
        log_dict: LogModel,
        log_prefix: str,
    ) -> AsyncGenerator[str, None]:
        if log_prefix:
            yield log_prefix

        is_stdio_log = log_dict.type == 's'

        async for line in self.master.db.logs.iter_log_lines(log_dict.id):
            if is_stdio_log:
                # for stdio logs, the first char is the stream type
                # ref: https://buildbot.readthedocs.io/en/latest/developer/raml/logchunk.html#logchunk
                line = line[1:]

            yield line

    @defer.inlineCallbacks
    def get_log_lines_raw_data(self, kwargs):
        retriever = base.NestedBuildDataRetriever(self.master, kwargs)
        log_dict = yield retriever.get_log_dict()
        if log_dict is None:
            return None, None, None

        # The following should be run sequentially instead of in gatherResults(), so that
        # they don't all start a query on step dict each.
        step_dict = yield retriever.get_step_dict()
        build_dict = yield retriever.get_build_dict()
        builder_dict = yield retriever.get_builder_dict()
        worker_dict = yield retriever.get_worker_dict()

        log_prefix = ''
        if log_dict.type == 's':
            if builder_dict is not None:
                log_prefix += f'Builder: {builder_dict.name}\n'
            if build_dict is not None:
                log_prefix += f'Build number: {build_dict.number}\n'
            if worker_dict is not None:
                log_prefix += f'Worker name: {worker_dict.name}\n'

        informative_parts = []
        if builder_dict is not None:
            informative_parts += [builder_dict.name]
        if build_dict is not None:
            informative_parts += ['build', str(build_dict.number)]
        if step_dict is not None:
            informative_parts += ['step', step_dict.name]
        informative_parts += ['log', log_dict.slug]
        informative_slug = '_'.join(informative_parts)

        return self.get_log_lines(log_dict, log_prefix), log_dict.type, informative_slug


class LogChunkEndpoint(LogChunkEndpointBase):
    # Note that this is a singular endpoint, even though it overrides the
    # offset/limit query params in ResultSpec
    kind = base.EndpointKind.SINGLE
    isPseudoCollection = True
    pathPatterns = [
        "/logchunks",
        "/logs/n:logid/contents",
        "/steps/n:stepid/logs/i:log_slug/contents",
        "/builds/n:buildid/steps/i:step_name/logs/i:log_slug/contents",
        "/builds/n:buildid/steps/n:step_number/logs/i:log_slug/contents",
        "/builders/n:builderid/builds/n:build_number/steps/i:step_name/logs/i:log_slug/contents",
        "/builders/n:builderid/builds/n:build_number/steps/n:step_number/logs/i:log_slug/contents",
    ]
    rootLinkName = "logchunks"

    @defer.inlineCallbacks
    def get(self, resultSpec, kwargs):
        retriever = base.NestedBuildDataRetriever(self.master, kwargs)
        logid = yield retriever.get_log_id()
        if logid is None:
            return None

        firstline = int(resultSpec.offset or 0)
        lastline = None if resultSpec.limit is None else firstline + int(resultSpec.limit) - 1
        resultSpec.removePagination()

        # get the number of lines, if necessary
        if lastline is None:
            log_dict = yield retriever.get_log_dict()
            if not log_dict:
                return None
            lastline = int(max(0, log_dict.num_lines - 1))

        # bounds checks
        if firstline < 0 or lastline < 0 or firstline > lastline:
            return None

        logLines = yield self.master.db.logs.getLogLines(logid, firstline, lastline)
        return {'logid': logid, 'firstline': firstline, 'content': logLines}


class RawLogChunkEndpoint(LogChunkEndpointBase):
    # Note that this is a singular endpoint, even though it overrides the
    # offset/limit query params in ResultSpec
    kind = base.EndpointKind.RAW
    pathPatterns = [
        "/logs/n:logid/raw",
        "/steps/n:stepid/logs/i:log_slug/raw",
        "/builds/n:buildid/steps/i:step_name/logs/i:log_slug/raw",
        "/builds/n:buildid/steps/n:step_number/logs/i:log_slug/raw",
        "/builders/n:builderid/builds/n:build_number/steps/i:step_name/logs/i:log_slug/raw",
        "/builders/n:builderid/builds/n:build_number/steps/n:step_number/logs/i:log_slug/raw",
    ]

    @defer.inlineCallbacks
    def get(self, resultSpec, kwargs):
        data = yield defer.Deferred.fromCoroutine(self.stream(resultSpec, kwargs))
        if data is None:
            return None

        data["raw"] = yield defer.Deferred.fromCoroutine(
            self.get_raw_log_lines(log_lines_generator=data["raw"])
        )
        return data

    async def stream(self, resultSpec: base.ResultSpec, kwargs: dict[str, Any]):
        log_lines_generator, log_type, log_slug = await self.get_log_lines_raw_data(kwargs)

        if log_lines_generator is None:
            return None

        return {
            'raw': log_lines_generator,
            'mime-type': 'text/html' if log_type == 'h' else 'text/plain',
            'filename': log_slug,
        }


class RawInlineLogChunkEndpoint(LogChunkEndpointBase):
    # Note that this is a singular endpoint, even though it overrides the
    # offset/limit query params in ResultSpec
    kind = base.EndpointKind.RAW_INLINE
    pathPatterns = [
        "/logs/n:logid/raw_inline",
        "/steps/n:stepid/logs/i:log_slug/raw_inline",
        "/builds/n:buildid/steps/i:step_name/logs/i:log_slug/raw_inline",
        "/builds/n:buildid/steps/n:step_number/logs/i:log_slug/raw_inline",
        "/builders/n:builderid/builds/n:build_number/steps/i:step_name/logs/i:log_slug/raw_inline",
        "/builders/n:builderid/builds/n:build_number/steps/n:step_number/logs/i:log_slug/raw_inline",
    ]

    @defer.inlineCallbacks
    def get(self, resultSpec, kwargs):
        data = yield defer.Deferred.fromCoroutine(self.stream(resultSpec, kwargs))

        if data is None:
            return None

        data["raw"] = yield defer.Deferred.fromCoroutine(
            self.get_raw_log_lines(log_lines_generator=data["raw"])
        )

        return data

    async def stream(self, resultSpec: base.ResultSpec, kwargs: dict[str, Any]):
        log_lines_generator, log_type, _ = await self.get_log_lines_raw_data(kwargs)

        if log_lines_generator is None:
            return None

        return {
            'raw': log_lines_generator,
            'mime-type': 'text/html' if log_type == 'h' else 'text/plain',
        }


class LogChunk(base.ResourceType):
    name = "logchunk"
    plural = "logchunks"
    endpoints = [LogChunkEndpoint, RawLogChunkEndpoint, RawInlineLogChunkEndpoint]

    class EntityType(types.Entity):
        logid = types.Integer()
        firstline = types.Integer()
        content = types.String()

    entityType = EntityType(name)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import asyncio
import json
import logging

import aiohttp  # dev-proxy command requires aiohttp! run 'pip install aiohttp'
import aiohttp.web
import jinja2

from buildbot.plugins.db import get_plugins

log = logging.getLogger(__name__)


class DevProxy:
    MAX_CONNECTIONS = 10

    def __init__(self, port, next_url, plugins, unsafe_ssl, auth_cookie):
        while next_url.endswith('/'):
            next_url = next_url[:-1]
        self.next_url = next_url
        self.app = app = aiohttp.web.Application()
        self.apps = get_plugins('www', None, load_now=True)
        self.unsafe_ssl = unsafe_ssl
        cookies = {}
        if auth_cookie:
            if "TWISTED_SESSION" in auth_cookie:  # user pasted the whole document.cookie part!
                cookies = dict(c.split("=") for c in auth_cookie.split(";"))
                auth_cookie = cookies["TWISTED_SESSION"]
            cookies = {'TWISTED_SESSION': auth_cookie}
        logging.basicConfig(level=logging.DEBUG)
        if plugins is None:
            plugins = {}
        else:
            plugins = json.loads(plugins)

        self.plugins = plugins

        app.router.add_route('*', '/ws', self.ws_handler)
        for path in ['/api', '/auth', '/sse', '/avatar']:
            app.router.add_route('*', path + '{path:.*}', self.proxy_handler)
        app.router.add_route('*', '/', self.index_handler)
        for plugin in self.apps.names:
            if plugin != 'base':
                staticdir = self.apps.get(plugin).static_dir
                app.router.add_static('/' + plugin, staticdir)
        staticdir = self.staticdir = self.apps.get('base').static_dir
        loader = jinja2.FileSystemLoader(staticdir)
        self.jinja = jinja2.Environment(loader=loader, undefined=jinja2.StrictUndefined)
        app.router.add_static('/', staticdir)
        conn = aiohttp.TCPConnector(limit=self.MAX_CONNECTIONS, verify_ssl=not self.unsafe_ssl)
        self.session = aiohttp.ClientSession(connector=conn, trust_env=True, cookies=cookies)
        self.config = None
        self.buildbotURL = f"http://localhost:{port}/"
        app.on_startup.append(self.on_startup)
        app.on_cleanup.append(self.on_cleanup)
        aiohttp.web.run_app(app, host="localhost", port=port)

    async def on_startup(self, app):
        try:
            await self.fetch_config_from_upstream()
        except aiohttp.ClientConnectionError as e:
            raise RuntimeError("Unable to connect to buildbot master" + str(e)) from e

    async def on_cleanup(self, app):
        await self.session.close()

    async def ws_handler(self, req):
        # based on https://github.com/oetiker/aio-reverse-proxy/blob/master/paraview-proxy.py
        ws_server = aiohttp.web.WebSocketResponse()
        await ws_server.prepare(req)

        async with self.session.ws_connect(self.next_url + "/ws", headers=req.headers) as ws_client:

            async def ws_forward(ws_from, ws_to):
                async for msg in ws_from:
                    if ws_to.closed:
                        await ws_to.close(code=ws_to.close_code, message=msg.extra)
                        return
                    if msg.type == aiohttp.WSMsgType.TEXT:
                        await ws_to.send_str(msg.data)
                    elif msg.type == aiohttp.WSMsgType.BINARY:
                        await ws_to.send_bytes(msg.data)
                    elif msg.type == aiohttp.WSMsgType.PING:
                        await ws_to.ping()
                    elif msg.type == aiohttp.WSMsgType.PONG:
                        await ws_to.pong()
                    else:
                        raise ValueError(f'unexpected message type: {msg}')

            # keep forwarding websocket data in both directions
            await asyncio.wait(
                [ws_forward(ws_server, ws_client), ws_forward(ws_client, ws_server)],
                return_when=asyncio.FIRST_COMPLETED,
            )
        return ws_server

    async def proxy_handler(self, req):
        method = getattr(self.session, req.method.lower())
        upstream_url = self.next_url + req.path
        headers = req.headers.copy()
        query = req.query
        try:
            # note that req.content is a StreamReader, so the data is streamed
            # and not fully loaded in memory (unlike with python-requests)
            async with method(
                upstream_url, headers=headers, params=query, allow_redirects=False, data=req.content
            ) as request:
                response = aiohttp.web.StreamResponse(
                    status=request.status, headers=request.headers
                )
                writer = await response.prepare(req)
                while True:
                    chunk = await request.content.readany()
                    if not chunk:
                        break
                    # using writer.write instead of response.write saves a few checks
                    await writer.write(chunk)
                return response
        except aiohttp.ClientConnectionError as e:
            return self.connection_error(e)

    def connection_error(self, error):
        return aiohttp.web.Response(
            text=f'Unable to connect to upstream server {self.next_url} ({error!s})', status=502
        )

    async def fetch_config_from_upstream(self):
        async with self.session.get(self.next_url) as request:
            index = await request.content.read()
            if request.status != 200:
                raise RuntimeError("Unable to fetch buildbot config: " + index.decode())
        # hack to parse the configjson from upstream buildbot config
        start_delimiter = b'angular.module("buildbot_config", []).constant("config", '
        start_index = index.index(start_delimiter)
        last_index = index.index(b')</script></html>')
        self.config = json.loads(index[start_index + len(start_delimiter) : last_index].decode())

        # keep the original config, but remove the plugins that we don't know
        for plugin in list(self.config['plugins'].keys()):
            if plugin not in self.apps:
                del self.config['plugins'][plugin]
                log.warn("warning: Missing plugin compared to original buildbot: %s", plugin)

        # add the plugins configs passed in cmdline
        for k, v in self.plugins.items():
            self.config['plugins'][k] = v

        self.config['buildbotURL'] = self.buildbotURL
        self.config['buildbotURLs'] = [self.buildbotURL, self.next_url + "/"]

    async def index_handler(self, req):
        tpl = self.jinja.get_template('index.html')
        index = tpl.render(
            configjson=json.dumps(self.config), custom_templates={}, config=self.config
        )
        return aiohttp.web.Response(body=index, content_type='text/html')


def devproxy(config):
    DevProxy(
        config['port'],
        config['buildbot_url'],
        config['plugins'],
        config['unsafe_ssl'],
        config['auth_cookie'],
    )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import os
import shutil
import weakref
from unittest import mock

from twisted.internet import defer
from twisted.internet import reactor

from buildbot.config.master import DBConfig as MasterDBConfig
from buildbot.config.master import MasterConfig
from buildbot.secrets.manager import SecretManager
from buildbot.test import fakedb
from buildbot.test.fake import bworkermanager
from buildbot.test.fake import fakedata
from buildbot.test.fake import fakemq
from buildbot.test.fake import msgmanager
from buildbot.test.fake import pbmanager
from buildbot.test.fake.botmaster import FakeBotMaster
from buildbot.test.fake.machine import FakeMachineManager
from buildbot.test.fake.secrets import FakeSecretStorage
from buildbot.test.util.db import resolve_test_db_url
from buildbot.util import service
from buildbot.util.twisted import async_to_deferred


class FakeCache:
    """Emulate an L{AsyncLRUCache}, but without any real caching.  This
    I{does} do the weakref part, to catch un-weakref-able objects."""

    def __init__(self, name, miss_fn):
        self.name = name
        self.miss_fn = miss_fn

    def get(self, key, **kwargs):
        d = self.miss_fn(key, **kwargs)

        @d.addCallback
        def mkref(x):
            if x is not None:
                weakref.ref(x)
            return x

        return d

    def put(self, key, val):
        pass


class FakeCaches:
    def get_cache(self, name, miss_fn):
        return FakeCache(name, miss_fn)


class FakeBuilder:
    def __init__(self, master=None, buildername="Builder"):
        if master:
            self.master = master
            self.botmaster = master.botmaster
        self.name = buildername


class FakeLogRotation:
    rotateLength = 42
    maxRotatedFiles = 42


class FakeMaster(service.MasterService):
    """
    Create a fake Master instance: a Mock with some convenience
    implementations:

    - Non-caching implementation for C{self.caches}
    """

    buildbotURL: str
    db: fakedb.FakeDBConnector
    mq: fakemq.FakeMQConnector
    data: fakedata.FakeDataConnector
    www: mock.Mock
    _test_want_db: bool = False
    _test_did_shutdown: bool = False

    def __init__(self, reactor, basedir='basedir', master_id=fakedb.FakeDBConnector.MASTER_ID):
        super().__init__()
        self._master_id = master_id
        self.reactor = reactor
        self.objectids = {}
        self.config = MasterConfig()
        self.caches = FakeCaches()
        self.pbmanager = pbmanager.FakePBManager()
        self.initLock = defer.DeferredLock()
        self.basedir = basedir
        self.botmaster = FakeBotMaster()
        self.botmaster.setServiceParent(self)
        self.name = 'fake:/master'
        self.httpservice = None
        self.masterid = master_id
        self.msgmanager = msgmanager.FakeMsgManager()
        self.workers = bworkermanager.FakeWorkerManager()
        self.workers.setServiceParent(self)
        self.machine_manager = FakeMachineManager()
        self.machine_manager.setServiceParent(self)
        self.log_rotation = FakeLogRotation()
        self.db = mock.Mock()
        self.next_objectid = 0
        self.config_version = 0

        def getObjectId(sched_name, class_name):
            k = (sched_name, class_name)
            try:
                rv = self.objectids[k]
            except KeyError:
                rv = self.objectids[k] = self.next_objectid
                self.next_objectid += 1
            return defer.succeed(rv)

        self.db.state.getObjectId = getObjectId

    def getObjectId(self):
        return defer.succeed(self._master_id)

    def subscribeToBuildRequests(self, callback):
        pass

    @defer.inlineCallbacks
    def stopService(self):
        yield super().stopService()
        yield self.test_shutdown()

    @defer.inlineCallbacks
    def test_shutdown(self):
        if self._test_did_shutdown:
            return
        self._test_did_shutdown = True
        if self._test_want_db:
            yield self.db._shutdown()
        if os.path.isdir(self.basedir):
            shutil.rmtree(self.basedir)


# Leave this alias, in case we want to add more behavior later


@async_to_deferred
async def make_master(
    testcase,
    wantMq=False,
    wantDb=False,
    wantData=False,
    wantRealReactor=False,
    wantGraphql=False,
    with_secrets: dict | None = None,
    url=None,
    db_url=None,
    sqlite_memory=True,
    auto_upgrade=True,
    auto_shutdown=True,
    check_version=True,
    auto_clean=True,
    **kwargs,
) -> FakeMaster:
    if wantRealReactor:
        _reactor = reactor
    else:
        assert testcase is not None, "need testcase for fake reactor"
        # The test case must inherit from TestReactorMixin and setup it.
        _reactor = testcase.reactor

    master = FakeMaster(_reactor, **kwargs)
    if url:
        master.buildbotURL = url
    if wantData:
        wantMq = wantDb = True
    if wantMq:
        assert testcase is not None, "need testcase for wantMq"
        master.mq = fakemq.FakeMQConnector(testcase)
        await master.mq.setServiceParent(master)
    if wantDb:
        assert testcase is not None, "need testcase for wantDb"
        master.db = fakedb.FakeDBConnector(
            master.basedir,
            testcase,
            auto_upgrade=auto_upgrade,
            check_version=check_version,
            auto_clean=auto_clean,
        )
        master._test_want_db = True

        if auto_shutdown:
            # Add before setup so that failed database setup would still be closed and wouldn't
            # affect further tests
            testcase.addCleanup(master.test_shutdown)

        master.db.configured_db_config = MasterDBConfig(resolve_test_db_url(db_url, sqlite_memory))
        if not os.path.exists(master.basedir):
            os.makedirs(master.basedir)
        await master.db.set_master(master)
        await master.db.setup()

    if wantData:
        master.data = fakedata.FakeDataConnector(master, testcase)

    if with_secrets is not None:
        secret_service = SecretManager()
        secret_service.services = [FakeSecretStorage(secretdict=with_secrets)]
        # This should be awaited, but no other call to `setServiceParent` are awaited here
        await secret_service.setServiceParent(master)

    return master



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from twisted.internet import defer
from twisted.python.reflect import namedObject

from buildbot.util import service


class MQConnector(service.ReconfigurableServiceMixin, service.AsyncMultiService):
    classes = {
        'simple': {
            'class': "buildbot.mq.simple.SimpleMQ",
            'keys': set(['debug']),
        },
        'wamp': {
            'class': "buildbot.mq.wamp.WampMQ",
            'keys': set(["router_url", "realm", "wamp_debug_level"]),
        },
    }
    name: str | None = 'mq'  # type: ignore[assignment]

    def __init__(self):
        super().__init__()
        self.impl = None  # set in setup
        self.impl_type = None  # set in setup

    @defer.inlineCallbacks
    def setup(self):
        assert not self.impl

        # imports are done locally so that we don't try to import
        # implementation-specific modules unless they're required.
        typ = self.master.config.mq['type']
        assert typ in self.classes  # this is checked by MasterConfig
        self.impl_type = typ
        cls = namedObject(self.classes[typ]['class'])
        self.impl = cls()

        # set up the impl as a child service
        yield self.impl.setServiceParent(self)

        # configure it (early)
        self.impl.reconfigServiceWithBuildbotConfig(self.master.config)

        # copy the methods onto this object for ease of access
        self.produce = self.impl.produce
        self.startConsuming = self.impl.startConsuming
        self.waitUntilEvent = self.impl.waitUntilEvent

    def reconfigServiceWithBuildbotConfig(self, new_config):
        # double-check -- the master ensures this in config checks
        assert self.impl_type == new_config.mq['type']

        return super().reconfigServiceWithBuildbotConfig(new_config)

    def produce(self, routing_key, data):
        # will be patched after configuration to point to the running
        # implementation's method
        raise NotImplementedError

    def startConsuming(self, callback, filter, persistent_name=None):
        # will be patched after configuration to point to the running
        # implementation's method
        raise NotImplementedError



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer

from buildbot.process import results
from buildbot.process.buildstep import BuildStep
from buildbot.process.buildstep import CommandMixin
from buildbot.test.util.integration import RunMasterBase


# This integration test creates a master and worker environment,
# and makes sure the command mixin is working.
class CommandMixinMaster(RunMasterBase):
    @defer.inlineCallbacks
    def setup_config(self):
        c = {}
        from buildbot.config import BuilderConfig
        from buildbot.plugins import schedulers
        from buildbot.process.factory import BuildFactory

        c['schedulers'] = [schedulers.AnyBranchScheduler(name="sched", builderNames=["testy"])]

        f = BuildFactory()
        f.addStep(TestCommandMixinStep())
        c['builders'] = [BuilderConfig(name="testy", workernames=["local1"], factory=f)]
        yield self.setup_master(c)

    @defer.inlineCallbacks
    def test_commandmixin(self):
        yield self.setup_config()

        change = {
            "branch": "master",
            "files": ["foo.c"],
            "author": "me@foo.com",
            "committer": "me@foo.com",
            "comments": "good stuff",
            "revision": "HEAD",
            "project": "none",
        }
        build = yield self.doForceBuild(wantSteps=True, useChange=change, wantLogs=True)
        self.assertEqual(build['buildid'], 1)
        self.assertEqual(build['results'], results.SUCCESS)


class CommandMixinMasterPB(CommandMixinMaster):
    proto = "pb"


class CommandMixinMasterMsgPack(CommandMixinMaster):
    proto = "msgpack"


class TestCommandMixinStep(BuildStep, CommandMixin):
    @defer.inlineCallbacks
    def run(self):
        contents = yield self.runGlob('*')
        if contents != []:
            return results.FAILURE

        hasPath = yield self.pathExists('composite_mixin_test')
        if hasPath:
            return results.FAILURE

        yield self.runMkdir('composite_mixin_test')

        hasPath = yield self.pathExists('composite_mixin_test')
        if not hasPath:
            return results.FAILURE

        contents = yield self.runGlob('*')
        if not contents[0].endswith('composite_mixin_test'):
            return results.FAILURE

        yield self.runRmdir('composite_mixin_test')

        hasPath = yield self.pathExists('composite_mixin_test')
        if hasPath:
            return results.FAILURE

        return results.SUCCESS



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer
from twisted.python import log

from buildbot.util import Notifier


class SubscriptionPoint:
    def __init__(self, name):
        self.name = name
        self.subscriptions = set()
        self._unfinished_deliveries = []
        self._unfinished_notifier = Notifier()
        self._got_exceptions = []

    def __str__(self):
        return f"<SubscriptionPoint '{self.name}'>"

    def subscribe(self, callback):
        sub = Subscription(self, callback)
        self.subscriptions.add(sub)
        return sub

    def deliver(self, *args, **kwargs):
        self._unfinished_deliveries.append(self)
        for sub in list(self.subscriptions):
            try:
                d = sub.callback(*args, **kwargs)
                if isinstance(d, defer.Deferred):
                    self._unfinished_deliveries.append(d)
                    d.addErrback(self._notify_delivery_exception, sub)
                    d.addBoth(self._notify_delivery_finished, d)

            except Exception as e:
                self._notify_delivery_exception(e, sub)

        self._notify_delivery_finished(None, self)

    def waitForDeliveriesToFinish(self):
        # returns a deferred
        if not self._unfinished_deliveries:
            return defer.succeed(None)
        return self._unfinished_notifier.wait()

    def pop_exceptions(self):
        exceptions = self._got_exceptions
        self._got_exceptions = None  # we no longer expect any exceptions
        return exceptions

    def _unsubscribe(self, subscription):
        self.subscriptions.remove(subscription)

    def _notify_delivery_exception(self, e, sub):
        log.err(e, f'while invoking callback {sub.callback} to {self}')
        if self._got_exceptions is None:
            log.err(
                e,
                'exceptions have already been collected. '
                'This is serious error, please submit a bug report',
            )
            return
        self._got_exceptions.append(e)

    def _notify_delivery_finished(self, _, d):
        self._unfinished_deliveries.remove(d)
        if not self._unfinished_deliveries:
            self._unfinished_notifier.notify(None)


class Subscription:
    def __init__(self, subpt, callback):
        self.subpt = subpt
        self.callback = callback

    def unsubscribe(self):
        self.subpt._unsubscribe(self)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from packaging.version import parse as parse_version
from twisted import __version__ as twistedVersion
from twisted.internet import defer

from buildbot.test.util.decorators import skipUnlessPlatformIs
from buildbot.test.util.integration import RunMasterBase


# This integration test creates a master and worker environment,
# with one builder and a shellcommand step, which use usePTY
class ShellMaster(RunMasterBase):
    @defer.inlineCallbacks
    def setup_config(self, usePTY):
        c = {}
        from buildbot.config import BuilderConfig
        from buildbot.plugins import schedulers
        from buildbot.plugins import steps
        from buildbot.process.factory import BuildFactory

        c['schedulers'] = [schedulers.ForceScheduler(name="force", builderNames=["testy"])]

        f = BuildFactory()
        f.addStep(
            steps.ShellCommand(
                command='if [ -t 1 ] ; then echo in a terminal; else echo "not a terminal"; fi',
                usePTY=usePTY,
            )
        )
        c['builders'] = [BuilderConfig(name="testy", workernames=["local1"], factory=f)]
        yield self.setup_master(c)

    @skipUnlessPlatformIs('posix')
    @defer.inlineCallbacks
    def test_usePTY(self):
        yield self.setup_config(usePTY=True)

        build = yield self.doForceBuild(wantSteps=True, wantLogs=True)
        self.assertEqual(build['buildid'], 1)
        res = yield self.checkBuildStepLogExist(build, "in a terminal", onlyStdout=True)
        self.assertTrue(res)

        # Twisted versions less than 17.1.0 would issue a warning:
        #    "Argument strings and environment keys/values passed to reactor.spawnProcess
        #    "should be str, not unicode."
        # This warning was unnecessary. Even in the old versions of Twisted, the
        # unicode arguments were encoded.  This warning was removed in Twisted here:
        #
        # https://github.com/twisted/twisted/commit/23fa3cc05549251ea4118e4e03354d58df87eaaa
        if parse_version(twistedVersion) < parse_version("17.1.0"):
            self.flushWarnings()

    @skipUnlessPlatformIs('posix')
    @defer.inlineCallbacks
    def test_NOusePTY(self):
        yield self.setup_config(usePTY=False)

        build = yield self.doForceBuild(wantSteps=True, wantLogs=True)
        self.assertEqual(build['buildid'], 1)
        res = yield self.checkBuildStepLogExist(build, "not a terminal", onlyStdout=True)
        self.assertTrue(res)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import base64
import subprocess
import time
from unittest.case import SkipTest

from parameterized import parameterized
from twisted.internet import defer

from buildbot.process.properties import Interpolate
from buildbot.secrets.providers.vault_hvac import HashiCorpVaultKvSecretProvider
from buildbot.secrets.providers.vault_hvac import VaultAuthenticatorToken
from buildbot.steps.shell import ShellCommand
from buildbot.test.util.decorators import skipUnlessPlatformIs
from buildbot.test.util.integration import RunMasterBase

# This integration test creates a master and worker environment,
# with one builders and a shellcommand step


# Test needs to be explicitly disabled on Windows, as docker may be present there, but not able
# to properly launch images.
@skipUnlessPlatformIs('posix')
class TestVaultHvac(RunMasterBase):
    @defer.inlineCallbacks
    def setup_config(self, secret_specifier):
        c = {}
        from buildbot.config import BuilderConfig
        from buildbot.plugins import schedulers
        from buildbot.process.factory import BuildFactory

        c['schedulers'] = [schedulers.ForceScheduler(name="force", builderNames=["testy"])]

        # note that as of August 2021, the vault docker image default to kv
        # version 2 to be enabled by default
        c['secretsProviders'] = [
            HashiCorpVaultKvSecretProvider(
                authenticator=VaultAuthenticatorToken('my_vaulttoken'),
                vault_server="http://localhost:8200",
                secrets_mount="secret",
            )
        ]

        f = BuildFactory()
        f.addStep(ShellCommand(command=Interpolate(f'echo {secret_specifier} | base64')))

        c['builders'] = [BuilderConfig(name="testy", workernames=["local1"], factory=f)]

        yield self.setup_master(c)

    def start_container(self, image_tag):
        try:
            image = f'vault:{image_tag}'
            subprocess.check_call(['docker', 'pull', image], stdout=subprocess.DEVNULL)

            subprocess.check_call(
                [
                    'docker',
                    'run',
                    '-d',
                    '-e',
                    'SKIP_SETCAP=yes',
                    '-e',
                    'VAULT_DEV_ROOT_TOKEN_ID=my_vaulttoken',
                    '-e',
                    'VAULT_TOKEN=my_vaulttoken',
                    '--name=vault_for_buildbot',
                    '-p',
                    '8200:8200',
                    image,
                ],
                stdout=subprocess.DEVNULL,
            )
            time.sleep(1)  # the container needs a little time to setup itself
            self.addCleanup(self.remove_container)

            subprocess.check_call(
                [
                    'docker',
                    'exec',
                    '-e',
                    'VAULT_ADDR=http://127.0.0.1:8200/',
                    'vault_for_buildbot',
                    'vault',
                    'kv',
                    'put',
                    'secret/key',
                    'value=word',
                ],
                stdout=subprocess.DEVNULL,
            )

            subprocess.check_call(
                [
                    'docker',
                    'exec',
                    '-e',
                    'VAULT_ADDR=http://127.0.0.1:8200/',
                    'vault_for_buildbot',
                    'vault',
                    'kv',
                    'put',
                    'secret/anykey',
                    'anyvalue=anyword',
                ],
                stdout=subprocess.DEVNULL,
            )

            subprocess.check_call(
                [
                    'docker',
                    'exec',
                    '-e',
                    'VAULT_ADDR=http://127.0.0.1:8200/',
                    'vault_for_buildbot',
                    'vault',
                    'kv',
                    'put',
                    'secret/key1/key2',
                    'id=val',
                ],
                stdout=subprocess.DEVNULL,
            )
        except (FileNotFoundError, subprocess.CalledProcessError) as e:
            raise SkipTest("Vault integration needs docker environment to be setup") from e

    def remove_container(self):
        subprocess.call(['docker', 'rm', '-f', 'vault_for_buildbot'], stdout=subprocess.DEVNULL)

    @defer.inlineCallbacks
    def do_secret_test(self, image_tag, secret_specifier, expected_obfuscation, expected_value):
        self.start_container(image_tag)
        yield self.setup_config(secret_specifier=secret_specifier)
        build = yield self.doForceBuild(wantSteps=True, wantLogs=True)
        self.assertEqual(build['buildid'], 1)

        patterns = [
            f"echo {expected_obfuscation}",
            base64.b64encode((expected_value + "\n").encode('utf-8')).decode('utf-8'),
        ]

        res = yield self.checkBuildStepLogExist(build, patterns)
        self.assertTrue(res)

    all_tags = [
        ('1.9.7',),
        ('1.10.5',),
        ('1.11.1',),
    ]

    @parameterized.expand(all_tags)
    @defer.inlineCallbacks
    def test_key(self, image_tag):
        yield self.do_secret_test(image_tag, '%(secret:key|value)s', '<key|value>', 'word')

    @parameterized.expand(all_tags)
    @defer.inlineCallbacks
    def test_key_any_value(self, image_tag):
        yield self.do_secret_test(
            image_tag, '%(secret:anykey|anyvalue)s', '<anykey|anyvalue>', 'anyword'
        )

    @parameterized.expand(all_tags)
    @defer.inlineCallbacks
    def test_nested_key(self, image_tag):
        yield self.do_secret_test(image_tag, '%(secret:key1/key2|id)s', '<key1/key2|id>', 'val')



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import importlib
import inspect
import os
import warnings

import twisted
from packaging.version import parse as parse_version
from twisted.trial import unittest
from twisted.trial.unittest import SkipTest
from zope.interface.verify import verifyClass

from buildbot.interfaces import IBuildStep
from buildbot.interfaces import IChangeSource
from buildbot.interfaces import IScheduler
from buildbot.interfaces import IWorker
from buildbot.plugins.db import get_plugins


def get_python_module_contents(package_name):
    spec = importlib.util.find_spec(package_name)
    if spec is None or spec.origin is None:
        return set()

    pathname = os.path.dirname(spec.origin)
    result = set()

    with os.scandir(pathname) as dir_entries:
        for dir_entry in dir_entries:
            filename = dir_entry.name

            if filename.startswith('__'):
                continue
            next_package_name = '.'.join((package_name, filename.partition('.')[0]))

            if dir_entry.is_file() and filename.endswith('.py'):
                result.add(next_package_name)

            if dir_entry.is_dir() and not filename.startswith('.'):
                # Ignore hidden directories added by various tooling an user may have, e.g.
                # .ropeproject
                result.add(next_package_name)
                result |= get_python_module_contents(next_package_name)

    return result


# NOTE: when running this test locally, make sure to reinstall master after every change to pick up
# new entry points.
class TestSetupPyEntryPoints(unittest.TestCase):
    def test_changes(self):
        known_not_exported = {
            'buildbot.changes.gerritchangesource.GerritChangeSourceBase',
            'buildbot.changes.base.ReconfigurablePollingChangeSource',
            'buildbot.changes.base.ChangeSource',
        }
        self.verify_plugins_registered(
            'changes', 'buildbot.changes', IChangeSource, known_not_exported
        )

    def test_schedulers(self):
        known_not_exported = {
            'buildbot.schedulers.basic.BaseBasicScheduler',
            'buildbot.schedulers.timed.Timed',
            'buildbot.schedulers.trysched.TryBase',
            'buildbot.schedulers.base.BaseScheduler',
            'buildbot.schedulers.base.ReconfigurableBaseScheduler',
            'buildbot.schedulers.timed.NightlyBase',
            'buildbot.schedulers.basic.Scheduler',
        }
        self.verify_plugins_registered(
            'schedulers', 'buildbot.schedulers', IScheduler, known_not_exported
        )

    def test_steps(self):
        known_not_exported = {
            'buildbot.steps.download_secret_to_worker.RemoveWorkerFileSecret',
            'buildbot.steps.source.base.Source',
            'buildbot.steps.download_secret_to_worker.DownloadSecretsToWorker',
            'buildbot.steps.worker.WorkerBuildStep',
            'buildbot.steps.vstudio.VisualStudio',
        }
        self.verify_plugins_registered('steps', 'buildbot.steps', IBuildStep, known_not_exported)

    def test_util(self):
        # work around Twisted bug 9384.
        if parse_version(twisted.__version__) < parse_version("18.9.0"):
            raise SkipTest('manhole.py can not be imported on old twisted and new python')

        known_not_exported = {
            'buildbot.util._notifier.Notifier',
            'buildbot.util.backoff.ExponentialBackoffEngineAsync',
            'buildbot.util.backoff.ExponentialBackoffEngineSync',
            'buildbot.util.backoff.BackoffTimeoutExceededError',
            'buildbot.util.backoff.ExponentialBackoffEngine',
            'buildbot.util.bbcollections.KeyedSets',
            'buildbot.util.codebase.AbsoluteSourceStampsMixin',
            'buildbot.util.config.ConfiguredMixin',
            'buildbot.util.debounce.Debouncer',
            'buildbot.util.deferwaiter.DeferWaiter',
            "buildbot.util.deferwaiter.NonRepeatedActionHandler",
            'buildbot.util.deferwaiter.RepeatedActionHandler',
            'buildbot.util.git.GitMixin',
            'buildbot.util.git.GitStepMixin',
            'buildbot.util.git.GitServiceAuth',
            'buildbot.util.git.AbstractGitAuth',
            'buildbot.util.git.GitStepAuth',
            'buildbot.util.giturlparse.GitUrl',
            'buildbot.util.httpclientservice.HTTPClientService',
            'buildbot.util.httpclientservice.HTTPSession',
            'buildbot.util.httpclientservice.TreqResponseWrapper',
            'buildbot.util.httpclientservice.TxRequestsResponseWrapper',
            'buildbot.util.kubeclientservice.KubeClientService',
            'buildbot.util.kubeclientservice.KubeConfigLoaderBase',
            'buildbot.util.latent.CompatibleLatentWorkerMixin',
            'buildbot.util.lineboundaries.LineBoundaryFinder',
            'buildbot.util.lru.AsyncLRUCache',
            'buildbot.util.lru.LRUCache',
            'buildbot.util.maildir.MaildirService',
            'buildbot.util.maildir.NoSuchMaildir',
            'buildbot.util.netstrings.NetstringParser',
            'buildbot.util.netstrings.NullAddress',
            'buildbot.util.netstrings.NullTransport',
            'buildbot.util.pathmatch.Matcher',
            'buildbot.util.poll.Poller',
            'buildbot.util.private_tempdir.PrivateTemporaryDirectory',
            'buildbot.util.protocol.LineBuffer',
            'buildbot.util.protocol.LineProcessProtocol',
            'buildbot.util.pullrequest.PullRequestMixin',
            'buildbot.util.queue.ConnectableThreadQueue',
            'buildbot.util.queue.UndoableQueue',
            'buildbot.util.raml.RamlLoader',
            'buildbot.util.raml.RamlSpec',
            'buildbot.util.runprocess.RunProcessPP',
            'buildbot.util.runprocess.RunProcess',
            'buildbot.util.sautils.InsertFromSelect',
            'buildbot.util.service.AsyncMultiService',
            'buildbot.util.service.AsyncService',
            'buildbot.util.service.BuildbotService',
            'buildbot.util.service.BuildbotServiceManager',
            'buildbot.util.service.ClusteredBuildbotService',
            'buildbot.util.service.IndependentAsyncMultiService',
            'buildbot.util.service.MasterService',
            'buildbot.util.service.ReconfigurableServiceMixin',
            'buildbot.util.service.SharedService',
            'buildbot.util.state.StateMixin',
            'buildbot.util.subscription.Subscription',
            'buildbot.util.subscription.SubscriptionPoint',
            'buildbot.util.test_result_submitter.TestResultSubmitter',
            "buildbot.util.watchdog.Watchdog",
            "buildbot.util.twisted.ThreadPool",
        }
        self.verify_plugins_registered('util', 'buildbot.util', None, known_not_exported)

    def test_reporters(self):
        known_not_exported = {
            'buildbot.reporters.base.ReporterBase',
            'buildbot.reporters.generators.utils.BuildStatusGeneratorMixin',
            'buildbot.reporters.gerrit.DEFAULT_REVIEW',
            'buildbot.reporters.gerrit.DEFAULT_SUMMARY',
            'buildbot.reporters.gerrit.GerritBuildEndStatusGenerator',
            'buildbot.reporters.gerrit.GerritBuildSetStatusGenerator',
            'buildbot.reporters.gerrit.GerritBuildStartStatusGenerator',
            'buildbot.reporters.gerrit.GerritStatusGeneratorBase',
            'buildbot.reporters.irc.IRCChannel',
            'buildbot.reporters.irc.IRCContact',
            'buildbot.reporters.irc.IrcStatusBot',
            'buildbot.reporters.irc.IrcStatusFactory',
            'buildbot.reporters.irc.UsageError',
            'buildbot.reporters.mail.Domain',
            'buildbot.reporters.message.MessageFormatterBase',
            'buildbot.reporters.message.MessageFormatterBaseJinja',
            'buildbot.reporters.telegram.TelegramChannel',
            'buildbot.reporters.telegram.TelegramContact',
            'buildbot.reporters.telegram.TelegramPollingBot',
            'buildbot.reporters.telegram.TelegramStatusBot',
            'buildbot.reporters.telegram.TelegramWebhookBot',
            'buildbot.reporters.words.Channel',
            'buildbot.reporters.words.Contact',
            'buildbot.reporters.words.ForceOptions',
            'buildbot.reporters.words.StatusBot',
            'buildbot.reporters.words.ThrottledClientFactory',
            'buildbot.reporters.words.UsageError',
            'buildbot.reporters.words.WebhookResource',
        }
        self.verify_plugins_registered('reporters', 'buildbot.reporters', None, known_not_exported)

    def test_secrets(self):
        known_not_exported = {
            'buildbot.secrets.manager.SecretManager',
            'buildbot.secrets.providers.base.SecretProviderBase',
            'buildbot.secrets.secret.SecretDetails',
            'buildbot.secrets.providers.vault_hvac.VaultAuthenticator',
        }
        self.verify_plugins_registered('secrets', 'buildbot.secrets', None, known_not_exported)

    def test_webhooks(self):
        # in the case of webhooks the entry points list modules, not classes, so
        # verify_plugins_registered won't work. For now let's ignore this edge case
        get_plugins('webhooks', None, load_now=True)

    def test_workers(self):
        known_not_exported = {
            'buildbot.worker.upcloud.UpcloudLatentWorker',
            'buildbot.worker.base.AbstractWorker',
            'buildbot.worker.latent.AbstractLatentWorker',
            'buildbot.worker.latent.LocalLatentWorker',
            'buildbot.worker.marathon.MarathonLatentWorker',
            'buildbot.worker.docker.DockerBaseWorker',
        }
        self.verify_plugins_registered('worker', 'buildbot.worker', IWorker, known_not_exported)

    def verify_plugins_registered(
        self, plugin_type, module_name, interface, known_not_exported=None
    ):
        # This will verify whether we can load plugins, i.e. whether the entry points are valid.
        plugins = get_plugins(plugin_type, interface, load_now=True)

        # Now verify that are no unregistered plugins left.
        existing_classes = self.get_existing_classes(module_name, interface)

        exported_classes = {
            f'{plugins._get_entry(name)._entry.module}.{name}' for name in plugins.names
        }
        if known_not_exported is None:
            known_not_exported = set()

        not_exported_classes = existing_classes - exported_classes - known_not_exported
        self.assertEqual(not_exported_classes, set())
        self.assertEqual(known_not_exported - existing_classes, set())

    def class_provides_iface(self, interface, klass):
        try:
            verifyClass(interface, klass)
            return True
        except Exception:
            return False

    def get_existing_classes(self, module_name, interface):
        existing_modules = get_python_module_contents(module_name)
        existing_classes = set()

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            for existing_module in existing_modules:
                module = importlib.import_module(existing_module)
                for name, obj in inspect.getmembers(module):
                    if name.startswith('_'):
                        continue
                    if inspect.isclass(obj) and obj.__module__ == existing_module:
                        if interface is not None and not self.class_provides_iface(interface, obj):
                            continue
                        existing_classes.add(f'{existing_module}.{name}')
        return existing_classes



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any

from twisted.internet import defer

from buildbot.data import base
from buildbot.data import types

if TYPE_CHECKING:
    from buildbot.db.test_results import TestResultModel
    from buildbot.util.twisted import InlineCallbacksType


def _db2data(model: TestResultModel):
    return {
        'test_resultid': model.id,
        'builderid': model.builderid,
        'test_result_setid': model.test_result_setid,
        'test_name': model.test_name,
        'test_code_path': model.test_code_path,
        'line': model.line,
        'duration_ns': model.duration_ns,
        'value': model.value,
    }


class TestResultsEndpoint(base.Endpoint):
    kind = base.EndpointKind.COLLECTION
    pathPatterns = [
        "/test_result_sets/n:test_result_setid/results",
    ]

    @defer.inlineCallbacks
    def get(self, resultSpec, kwargs):
        set_dbdict = yield self.master.db.test_result_sets.getTestResultSet(
            kwargs['test_result_setid']
        )

        if set_dbdict is None:
            return []

        result_dbdicts = yield self.master.db.test_results.getTestResults(
            set_dbdict.builderid, kwargs['test_result_setid'], result_spec=resultSpec
        )

        return [_db2data(result) for result in result_dbdicts]


class TestResult(base.ResourceType):
    name = "test_result"
    plural = "test_results"
    endpoints = [TestResultsEndpoint]
    eventPathPatterns = [
        "/test_result_sets/:test_result_setid/results",
    ]

    class EntityType(types.Entity):
        test_resultid = types.Integer()
        builderid = types.Integer()
        test_result_setid = types.Integer()
        test_name = types.NoneOk(types.String())
        test_code_path = types.NoneOk(types.String())
        line = types.NoneOk(types.Integer())
        duration_ns = types.NoneOk(types.Integer())
        value = types.String()

    entityType = EntityType(name)

    @base.updateMethod
    @defer.inlineCallbacks
    def addTestResults(
        self, builderid: int, test_result_setid: int, result_values: list[dict[str, Any]]
    ) -> InlineCallbacksType[None]:
        # We're not adding support for emitting any messages, because in all cases all test results
        # will be part of a test result set. The users should wait for a 'complete' event on a
        # test result set and only then fetch the test results, which won't change from that time
        # onward.
        yield self.master.db.test_results.addTestResults(
            builderid, test_result_setid, result_values
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

"""
A complete re-implementation of the database connector components, but without
using a database.  These classes should pass the same tests as are applied to
the real connector components.
"""

from .build_data import BuildData
from .builders import Builder
from .builders import BuilderMaster
from .builders import BuildersTags
from .buildrequests import BuildRequest
from .buildrequests import BuildRequestClaim
from .builds import Build
from .builds import BuildProperty
from .buildsets import Buildset
from .buildsets import BuildsetProperty
from .buildsets import BuildsetSourceStamp
from .changes import Change
from .changes import ChangeFile
from .changes import ChangeProperty
from .changes import ChangeUser
from .changesources import ChangeSource
from .changesources import ChangeSourceMaster
from .codebases import Codebase
from .codebases import CodebaseBranch
from .codebases import CodebaseCommit
from .connector import FakeDBConnector
from .logs import Log
from .logs import LogChunk
from .masters import Master
from .projects import Project
from .schedulers import Scheduler
from .schedulers import SchedulerChange
from .schedulers import SchedulerMaster
from .sourcestamps import Patch
from .sourcestamps import SourceStamp
from .state import Object
from .state import ObjectState
from .steps import Step
from .tags import Tag
from .test_result_sets import TestResultSet
from .test_results import TestCodePath
from .test_results import TestName
from .test_results import TestResult
from .users import User
from .users import UserInfo
from .workers import ConfiguredWorker
from .workers import ConnectedWorker
from .workers import Worker

__all__ = [
    'Build',
    'BuildData',
    'BuildProperty',
    'BuildRequest',
    'BuildRequestClaim',
    'Builder',
    'BuilderMaster',
    'BuildersTags',
    'Buildset',
    'BuildsetProperty',
    'BuildsetSourceStamp',
    'Change',
    'ChangeFile',
    'ChangeProperty',
    'ChangeSource',
    'ChangeSourceMaster',
    'ChangeUser',
    'Codebase',
    'CodebaseBranch',
    'CodebaseCommit',
    'ConfiguredWorker',
    'ConnectedWorker',
    'FakeDBConnector',
    'Log',
    'LogChunk',
    'Master',
    'Object',
    'ObjectState',
    'Patch',
    'Project',
    'Scheduler',
    'SchedulerChange',
    'SchedulerMaster',
    'SourceStamp',
    'Step',
    'Tag',
    'TestCodePath',
    'TestName',
    'TestResult',
    'TestResultSet',
    'User',
    'UserInfo',
    'Worker',
]



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

# This change hook allows GitHub or a hand crafted curl invocation to "knock on
# the door" and trigger a change source to poll.


from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any
from typing import cast

from buildbot.changes.base import ReconfigurablePollingChangeSource
from buildbot.util import bytes2unicode
from buildbot.util import unicode2bytes
from buildbot.www.hooks.base import BaseHookHandler
from buildbot.www.service import BuildbotSite

if TYPE_CHECKING:
    from twisted.internet import defer
    from twisted.web.server import Request


class PollingHandler(BaseHookHandler):
    def getChanges(self, req: Request) -> defer.Deferred[tuple[list[dict[str, Any]], str | None]]:
        site = cast(BuildbotSite, req.site)
        assert site.master is not None
        change_svc = site.master.change_svc
        assert req.args is not None
        poll_all = b"poller" not in req.args

        allow_all = True
        allowed: list[bytes] = []
        if isinstance(self.options, dict) and b"allowed" in self.options:
            allow_all = False
            allowed = self.options[b"allowed"]

        pollers: list[ReconfigurablePollingChangeSource] = []

        for source in change_svc:
            if not isinstance(source, ReconfigurablePollingChangeSource):
                continue
            if not hasattr(source, "name"):
                continue
            if not poll_all and unicode2bytes(source.name) not in req.args[b'poller']:
                continue
            if not allow_all and unicode2bytes(source.name) not in allowed:
                continue
            pollers.append(source)

        if not poll_all:
            missing = set(req.args[b'poller']) - set(unicode2bytes(s.name) for s in pollers)
            if missing:
                raise ValueError(f'Could not find pollers: {bytes2unicode(b",".join(missing))}')

        for p in pollers:
            p.force()

        return [], None


poller = PollingHandler



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import datetime
import locale

from twisted.python import log
from twisted.trial import unittest

from buildbot.test.util import validation
from buildbot.util import UTC


class VerifyDict(unittest.TestCase):
    def doValidationTest(self, validator, good, bad):
        for g in good:
            log.msg(f'expect {g!r} to be good')
            msgs = list(validator.validate('g', g))
            self.assertEqual(msgs, [], f'messages for {g!r}')
        for b in bad:
            log.msg(f'expect {b!r} to be bad')
            msgs = list(validator.validate('b', b))
            self.assertNotEqual(msgs, [], f'no messages for {b!r}')
            log.msg('..got messages:')
            for msg in msgs:
                log.msg("  " + msg)

    def test_IntValidator(self):
        self.doValidationTest(
            validation.IntValidator(), good=[1, 10**100], bad=[1.0, "one", "1", None]
        )

    def test_BooleanValidator(self):
        self.doValidationTest(
            validation.BooleanValidator(), good=[True, False], bad=["yes", "no", 1, 0, None]
        )

    def test_StringValidator(self):
        self.doValidationTest(
            validation.StringValidator(), good=["unicode only"], bad=[None, b"bytestring"]
        )

    def test_BinaryValidator(self):
        self.doValidationTest(
            validation.BinaryValidator(), good=[b"bytestring"], bad=[None, "no unicode"]
        )

    def test_DateTimeValidator(self):
        self.doValidationTest(
            validation.DateTimeValidator(),
            good=[
                datetime.datetime(1980, 6, 15, 12, 31, 15, tzinfo=UTC),
            ],
            bad=[
                None,
                198847493,
                # no timezone
                datetime.datetime(1980, 6, 15, 12, 31, 15),
            ],
        )

    def test_IdentifierValidator(self):
        os_encoding = locale.getpreferredencoding()
        try:
            '\N{SNOWMAN}'.encode(os_encoding)
        except UnicodeEncodeError as e:
            # Default encoding of Windows console is 'cp1252'
            # which cannot encode the snowman.
            raise unittest.SkipTest(
                f"Cannot encode weird unicode on this platform with {os_encoding}"
            ) from e

        self.doValidationTest(
            validation.IdentifierValidator(50),
            good=["linux", "Linux", "abc123", "a" * 50, '\N{SNOWMAN}'],
            bad=[
                None,
                '',
                b'linux',
                'a/b',
                "a.b.c.d",
                "a-b_c.d9",
                'spaces not allowed',
                "a" * 51,
                "123 no initial digits",
            ],
        )

    def test_NoneOk(self):
        self.doValidationTest(
            validation.NoneOk(validation.BooleanValidator()),
            good=[True, False, None],
            bad=[1, "yes"],
        )

    def test_DictValidator(self):
        self.doValidationTest(
            validation.DictValidator(
                a=validation.BooleanValidator(), b=validation.StringValidator(), optionalNames=['b']
            ),
            good=[
                {'a': True},
                {'a': True, 'b': 'xyz'},
            ],
            bad=[
                None,
                1,
                "hi",
                {},
                {'a': 1},
                {'a': 1, 'b': 'xyz'},
                {'a': True, 'b': 999},
                {'a': True, 'b': 'xyz', 'c': 'extra'},
            ],
        )

    def test_DictValidator_names(self):
        v = validation.DictValidator(a=validation.BooleanValidator())
        self.assertEqual(list(v.validate('v', {'a': 1})), ["v['a'] (1) is not a boolean"])

    def test_ListValidator(self):
        self.doValidationTest(
            validation.ListValidator(validation.BooleanValidator()),
            good=[
                [],
                [True],
                [False, True],
            ],
            bad=[None, ['a'], [True, 'a'], 1, "hi"],
        )

    def test_ListValidator_names(self):
        v = validation.ListValidator(validation.BooleanValidator())
        self.assertEqual(list(v.validate('v', ['a'])), ["v[0] ('a') is not a boolean"])

    def test_SourcedPropertiesValidator(self):
        self.doValidationTest(
            validation.SourcedPropertiesValidator(),
            good=[
                {'pname': ('{"a":"b"}', 'test')},
            ],
            bad=[
                None,
                1,
                b"hi",
                {'pname': {b'a': b'b'}},  # no source
                # name not unicode
                {'pname': ({b'a': b'b'}, 'test')},
                # source not unicode
                {'pname': ({b'a': b'b'}, 'test')},
                # self is not json-able
                {'pname': (self, 'test')},
            ],
        )

    def test_MessageValidator(self):
        self.doValidationTest(
            validation.MessageValidator(
                events=[b'started', b'stopped'],
                messageValidator=validation.DictValidator(
                    a=validation.BooleanValidator(),
                    xid=validation.IntValidator(),
                    yid=validation.IntValidator(),
                ),
            ),
            good=[
                (('thing', '1', '2', 'started'), {'xid': 1, 'yid': 2, 'a': True}),
            ],
            bad=[
                # routingKey is not a tuple
                ('thing', {}),
                # routingKey has wrong event
                (('thing', '1', '2', 'exploded'), {'xid': 1, 'yid': 2, 'a': True}),
                # routingKey element has wrong type
                (('thing', 1, 2, 'started'), {'xid': 1, 'yid': 2, 'a': True}),
                # routingKey element isn't in message
                (('thing', '1', '2', 'started'), {'xid': 1, 'a': True}),
                # message doesn't validate
                (('thing', '1', '2', 'started'), {'xid': 1, 'yid': 2, 'a': 'x'}),
            ],
        )

    def test_Selector(self):
        sel = validation.Selector()
        sel.add(lambda x: x == 'int', validation.IntValidator())
        sel.add(lambda x: x == 'str', validation.StringValidator())
        self.doValidationTest(
            sel,
            good=[
                ('int', 1),
                ('str', 'hi'),
            ],
            bad=[
                ('int', 'hi'),
                ('str', 1),
                ('float', 1.0),
            ],
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import os
import stat

from twisted.python.filepath import FilePath
from twisted.trial import unittest

from buildbot.process import remotetransfer
from buildbot.process.results import SUCCESS
from buildbot.steps.download_secret_to_worker import DownloadSecretsToWorker
from buildbot.steps.download_secret_to_worker import RemoveWorkerFileSecret
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.steps import ExpectDownloadFile
from buildbot.test.steps import ExpectRemoteRef
from buildbot.test.steps import ExpectRmdir
from buildbot.test.steps import ExpectRmfile
from buildbot.test.steps import TestBuildStepMixin
from buildbot.test.util import config as configmixin


class TestDownloadFileSecretToWorkerCommand(
    TestBuildStepMixin, TestReactorMixin, unittest.TestCase
):
    def setUp(self):
        self.setup_test_reactor()
        tempdir = FilePath(self.mktemp())
        tempdir.createDirectory()
        self.temp_path = tempdir.path
        return self.setup_test_build_step()

    def testBasic(self):
        self.setup_step(
            DownloadSecretsToWorker([
                (os.path.join(self.temp_path, "pathA"), "something"),
                (os.path.join(self.temp_path, "pathB"), "something more"),
            ])
        )
        self.expect_commands(
            ExpectDownloadFile(
                maxsize=None,
                mode=stat.S_IRUSR | stat.S_IWUSR,
                reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                blocksize=32 * 1024,
                workerdest=os.path.join(self.temp_path, "pathA"),
                workdir="wkdir",
            ).exit(0),
            ExpectDownloadFile(
                maxsize=None,
                mode=stat.S_IRUSR | stat.S_IWUSR,
                reader=ExpectRemoteRef(remotetransfer.StringFileReader),
                blocksize=32 * 1024,
                workerdest=os.path.join(self.temp_path, "pathB"),
                workdir="wkdir",
            ).exit(0),
        )

        self.expect_outcome(result=SUCCESS, state_string="finished")
        d = self.run_step()
        return d


class TestRemoveWorkerFileSecretCommand30(TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        tempdir = FilePath(self.mktemp())
        tempdir.createDirectory()
        self.temp_path = tempdir.path
        return self.setup_test_build_step()

    def testBasic(self):
        self.setup_build(worker_version={'*': '3.0'})
        self.setup_step(
            RemoveWorkerFileSecret([
                (os.path.join(self.temp_path, "pathA"), "something"),
                (os.path.join(self.temp_path, "pathB"), "somethingmore"),
            ]),
        )

        self.expect_commands(
            ExpectRmdir(
                path=os.path.join(self.temp_path, "pathA"),
                dir=os.path.abspath(os.path.join(self.temp_path, "pathA")),
                log_environ=False,
            ).exit(0),
            ExpectRmdir(
                path=os.path.join(self.temp_path, "pathB"),
                dir=os.path.abspath(os.path.join(self.temp_path, "pathB")),
                log_environ=False,
            ).exit(0),
        )

        self.expect_outcome(result=SUCCESS, state_string="finished")
        d = self.run_step()
        return d


class TestRemoveFileSecretToWorkerCommand(
    TestBuildStepMixin, configmixin.ConfigErrorsMixin, TestReactorMixin, unittest.TestCase
):
    def setUp(self):
        self.setup_test_reactor()
        tempdir = FilePath(self.mktemp())
        tempdir.createDirectory()
        self.temp_path = tempdir.path
        return self.setup_test_build_step()

    def testBasic(self):
        self.setup_step(
            RemoveWorkerFileSecret([
                (os.path.join(self.temp_path, "pathA"), "something"),
                (os.path.join(self.temp_path, "pathB"), "somethingmore"),
            ])
        )
        self.expect_commands(
            ExpectRmfile(path=os.path.join(self.temp_path, "pathA"), log_environ=False).exit(0),
            ExpectRmfile(path=os.path.join(self.temp_path, "pathB"), log_environ=False).exit(0),
        )

        self.expect_outcome(result=SUCCESS, state_string="finished")
        d = self.run_step()
        return d



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from unittest import mock

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.mq import simple
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin


class SimpleMQ(TestReactorMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self)
        self.mq = simple.SimpleMQ()
        self.mq.setServiceParent(self.master)
        yield self.mq.startService()

        @defer.inlineCallbacks
        def cleanup():
            if self.mq.running:
                yield self.mq.stopService()

        self.addCleanup(cleanup)

    @defer.inlineCallbacks
    def test_forward_data(self):
        callback = mock.Mock()
        yield self.mq.startConsuming(callback, ('a', 'b'))
        # _produce returns a deferred
        yield self.mq.produce(('a', 'b'), 'foo')
        # calling produce should eventually call the callback with decoding of
        # topic
        callback.assert_called_with(('a', 'b'), 'foo')

    @defer.inlineCallbacks
    def test_forward_data_wildcard(self):
        callback = mock.Mock()
        yield self.mq.startConsuming(callback, ('a', None))
        # _produce returns a deferred
        yield self.mq.produce(('a', 'b'), 'foo')
        # calling produce should eventually call the callback with decoding of
        # topic
        callback.assert_called_with(('a', 'b'), 'foo')

    @defer.inlineCallbacks
    def test_waits_for_called_callback(self):
        def callback(_, __):
            return defer.succeed(None)

        yield self.mq.startConsuming(callback, ('a', None))
        yield self.mq.produce(('a', 'b'), 'foo')

        d = self.mq.stopService()
        self.assertTrue(d.called)

    @defer.inlineCallbacks
    def test_waits_for_non_called_callback(self):
        d1 = defer.Deferred()

        def callback(_, __):
            return d1

        yield self.mq.startConsuming(callback, ('a', None))
        yield self.mq.produce(('a', 'b'), 'foo')

        d = self.mq.stopService()
        self.assertFalse(d.called)
        d1.callback(None)
        self.assertTrue(d.called)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

"""
module for regrouping all FileWriterImpl and FileReaderImpl away from steps
"""

import os
import shutil
import tarfile
import tempfile
from io import BytesIO

from buildbot.util import bytes2unicode
from buildbot.util import unicode2bytes
from buildbot.worker.protocols import base


class FileWriter(base.FileWriterImpl):
    """
    Helper class that acts as a file-object with write access
    """

    def __init__(self, destfile, maxsize, mode):
        # Create missing directories.
        destfile = os.path.abspath(destfile)
        dirname = os.path.dirname(destfile)
        if not os.path.exists(dirname):
            os.makedirs(dirname)

        self.destfile = destfile
        self.mode = mode
        fd, self.tmpname = tempfile.mkstemp(dir=dirname, prefix='buildbot-transfer-')
        self.fp = os.fdopen(fd, 'wb')
        self.remaining = maxsize

    def remote_write(self, data):
        """
        Called from remote worker to write L{data} to L{fp} within boundaries
        of L{maxsize}

        @type  data: C{string}
        @param data: String of data to write
        """
        data = unicode2bytes(data)
        if self.remaining is not None:
            if len(data) > self.remaining:
                data = data[: self.remaining]
            self.fp.write(data)
            self.remaining = self.remaining - len(data)
        else:
            self.fp.write(data)

    def remote_utime(self, accessed_modified):
        os.utime(self.destfile, accessed_modified)

    def remote_close(self):
        """
        Called by remote worker to state that no more data will be transferred
        """
        self.fp.close()
        self.fp = None
        # on windows, os.rename does not automatically unlink, so do it
        # manually
        if os.path.exists(self.destfile):
            os.unlink(self.destfile)
        os.rename(self.tmpname, self.destfile)
        self.tmpname = None
        if self.mode is not None:
            os.chmod(self.destfile, self.mode)

    def cancel(self):
        # unclean shutdown, the file is probably truncated, so delete it
        # altogether rather than deliver a corrupted file
        fp = getattr(self, "fp", None)
        if fp:
            fp.close()
            self.purge()

    def purge(self):
        if self.destfile and os.path.exists(self.destfile):
            os.unlink(self.destfile)
        if self.tmpname and os.path.exists(self.tmpname):
            os.unlink(self.tmpname)


class DirectoryWriter(FileWriter):
    """
    A DirectoryWriter is implemented as a FileWriter, with an added post-processing
    step to unpack the archive, once the transfer has completed.
    """

    def __init__(self, destroot, maxsize, compress, mode):
        self.destroot = destroot
        self.compress = compress

        self.fd, self.tarname = tempfile.mkstemp(prefix='buildbot-transfer-')
        os.close(self.fd)

        super().__init__(self.tarname, maxsize, mode)

    def remote_unpack(self):
        """
        Called by remote worker to state that no more data will be transferred
        """
        # Make sure remote_close is called, otherwise atomic rename won't happen
        self.remote_close()

        # Map configured compression to a TarFile setting
        if self.compress == 'bz2':
            mode = 'r|bz2'
        elif self.compress == 'gz':
            mode = 'r|gz'
        else:
            mode = 'r'

        # Unpack archive and clean up after self
        with tarfile.open(name=self.tarname, mode=mode) as archive:
            if hasattr(tarfile, 'data_filter'):
                archive.extractall(path=self.destroot, filter='data')
            else:
                archive.extractall(path=self.destroot)
        os.remove(self.tarname)

    def purge(self):
        super().purge()
        if os.path.isdir(self.destroot):
            shutil.rmtree(self.destroot)


class FileReader(base.FileReaderImpl):
    """
    Helper class that acts as a file-object with read access
    """

    def __init__(self, fp):
        self.fp = fp

    def remote_read(self, maxlength):
        """
        Called from remote worker to read at most L{maxlength} bytes of data

        @type  maxlength: C{integer}
        @param maxlength: Maximum number of data bytes that can be returned

        @return: Data read from L{fp}
        @rtype: C{string} of bytes read from file
        """
        if self.fp is None:
            return ''

        data = self.fp.read(maxlength)
        return data

    def remote_close(self):
        """
        Called by remote worker to state that no more data will be transferred
        """
        if self.fp is not None:
            self.fp.close()
            self.fp = None


class StringFileWriter(base.FileWriterImpl):
    """
    FileWriter class that just puts received data into a buffer.

    Used to upload a file from worker for inline processing rather than
    writing into a file on master.
    """

    def __init__(self):
        self.buffer = ""

    def remote_write(self, data):
        self.buffer += bytes2unicode(data)

    def remote_close(self):
        pass


class StringFileReader(FileReader):
    """
    FileWriter class that just buid send data from a string.

    Used to download a file to worker from local string rather than first
    writing into a file on master.
    """

    def __init__(self, s):
        s = unicode2bytes(s)
        super().__init__(BytesIO(s))



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import errno
import os
import signal
import time

from twisted.python.runtime import platformType

from buildbot.scripts import base


def stop(config, signame="TERM", wait=None):
    basedir = config['basedir']
    quiet = config['quiet']

    if wait is None:
        wait = not config['no-wait']

    if config['clean']:
        signame = 'USR1'

    if not base.isBuildmasterDir(config['basedir']):
        return 1

    pidfile = os.path.join(basedir, 'twistd.pid')
    try:
        with open(pidfile, encoding='utf-8') as f:
            pid = int(f.read().strip())
    except Exception:
        if not config['quiet']:
            print("buildmaster not running")
        return 0

    signum = getattr(signal, "SIG" + signame)
    try:
        os.kill(pid, signum)
    except OSError as e:
        if e.errno != errno.ESRCH and platformType != "win32":
            raise

        if not config['quiet']:
            print("buildmaster not running")
        try:
            os.unlink(pidfile)
        except OSError:
            pass
        return 0

    if not wait:
        if not quiet:
            print(f"sent SIG{signame} to process")
        return 0

    time.sleep(0.1)

    # poll once per second until twistd.pid goes away, up to 10 seconds,
    # unless we're doing a clean stop, in which case wait forever
    count = 0
    while count < 10 or config['clean']:
        try:
            os.kill(pid, 0)
        except OSError:
            if not quiet:
                print(f"buildbot process {pid} is dead")
            return 0
        time.sleep(1)
        count += 1
    if not quiet:
        print("never saw process go away")
    return 1



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from typing import ClassVar
from typing import Sequence

from twisted.internet import defer

from buildbot import config
from buildbot.data import resultspec
from buildbot.process.results import FAILURE
from buildbot.util.service import BuildbotService
from buildbot.util.ssfilter import SourceStampFilter
from buildbot.util.ssfilter import extract_filter_values


class _FailingSingleBuilderConfig:
    def __init__(self, builders_to_cancel, filter):
        self.builders_to_cancel = builders_to_cancel
        self.filter = filter


class _FailingBuilderConfig:
    def __init__(self):
        self._by_builder = {}

    def add_config(self, builders, builders_to_cancel, filter):
        assert builders is not None

        config = _FailingSingleBuilderConfig(builders_to_cancel, filter)

        for builder in builders:
            self._by_builder.setdefault(builder, []).append(config)

    def get_all_matched(self, builder_name, props):
        assert builder_name is not None

        configs = self._by_builder.get(builder_name, [])
        return [c for c in configs if c.filter.is_matched(props)]


class FailingBuildsetCanceller(BuildbotService):
    compare_attrs: ClassVar[Sequence[str]] = (*BuildbotService.compare_attrs, 'filters')

    def checkConfig(self, name, filters):
        FailingBuildsetCanceller.check_filters(filters)

        self.name = name

        self._build_finished_consumer = None

    def reconfigService(self, name, filters):
        self.filters = FailingBuildsetCanceller.filter_tuples_to_filter_set_object(filters)

    @defer.inlineCallbacks
    def startService(self):
        yield super().startService()
        self._build_finished_consumer = yield self.master.mq.startConsuming(
            self._on_build_finished, ('builds', None, 'finished')
        )

    @defer.inlineCallbacks
    def stopService(self):
        yield self._build_finished_consumer.stopConsuming()

    @classmethod
    def check_filters(cls, filters):
        if not isinstance(filters, list):
            config.error(f'{cls.__name__}: The filters argument must be a list of tuples')

        for filter in filters:
            if (
                not isinstance(filter, tuple)
                or len(filter) != 3
                or not isinstance(filter[2], SourceStampFilter)
            ):
                config.error(
                    (
                        '{}: The filters argument must be a list of tuples each of which '
                        + 'contains builders to track as the first item, builders to cancel '
                        + 'as the second and SourceStampFilter as the third'
                    ).format(cls.__name__)
                )

            builders, builders_to_cancel, _ = filter

            try:
                extract_filter_values(builders, 'builders')
                if builders_to_cancel is not None:
                    extract_filter_values(builders_to_cancel, 'builders_to_cancel')
            except Exception as e:
                config.error(f'{cls.__name__}: When processing filter builders: {e!s}')

    @classmethod
    def filter_tuples_to_filter_set_object(cls, filters):
        filter_set = _FailingBuilderConfig()

        for filter in filters:
            builders, builders_to_cancel, ss_filter = filter

            extract_filter_values(builders, 'builders')

            if builders_to_cancel is not None:
                builders_to_cancel = extract_filter_values(builders_to_cancel, 'builders_to_cancel')

            filter_set.add_config(builders, builders_to_cancel, ss_filter)

        return filter_set

    @defer.inlineCallbacks
    def _on_build_finished(self, key, build):
        if build['results'] != FAILURE:
            return

        buildrequest = yield self.master.data.get(('buildrequests', build['buildrequestid']))
        builder = yield self.master.data.get(("builders", build['builderid']))
        buildset = yield self.master.data.get(('buildsets', buildrequest['buildsetid']))

        sourcestamps = buildset['sourcestamps']

        builders_to_cancel = set()
        for ss in sourcestamps:
            configs = self.filters.get_all_matched(builder['name'], ss)
            for c in configs:
                if builders_to_cancel is not None:
                    if c.builders_to_cancel is None:
                        builders_to_cancel = None
                    else:
                        builders_to_cancel.update(c.builders_to_cancel)

        all_bs_buildrequests = yield self.master.data.get(
            ('buildrequests',),
            filters=[
                resultspec.Filter('buildsetid', 'eq', [buildset['bsid']]),
                resultspec.Filter('complete', 'eq', [False]),
            ],
        )

        all_bs_buildrequests = [
            br
            for br in all_bs_buildrequests
            if br['buildrequestid'] != buildrequest['buildrequestid']
        ]

        for br in all_bs_buildrequests:
            brid = br['buildrequestid']
            if brid == buildrequest['buildrequestid']:
                continue  # this one has just failed

            br_builder = yield self.master.data.get(("builders", br['builderid']))

            if builders_to_cancel is not None and br_builder['name'] not in builders_to_cancel:
                continue

            reason = 'Build has been cancelled because another build in the same buildset failed'

            self.master.data.control('cancel', {'reason': reason}, ('buildrequests', str(brid)))



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer

from buildbot.process.results import CANCELLED
from buildbot.test.util.decorators import flaky
from buildbot.test.util.integration import RunMasterBase
from buildbot.util import asyncSleep


class InterruptCommand(RunMasterBase):
    """Make sure we can interrupt a command"""

    @defer.inlineCallbacks
    def setup_config(self):
        c = {}
        from buildbot.plugins import schedulers
        from buildbot.plugins import steps
        from buildbot.plugins import util

        class SleepAndInterrupt(steps.ShellSequence):
            @defer.inlineCallbacks
            def run(self):
                if self.worker.worker_system == "nt":
                    sleep = "waitfor SomethingThatIsNeverHappening /t 100 >nul 2>&1"
                else:
                    sleep = ["sleep", "100"]
                d = self.runShellSequence([util.ShellArg(sleep)])
                yield asyncSleep(1)
                self.interrupt("just testing")
                res = yield d
                return res

        c['schedulers'] = [schedulers.ForceScheduler(name="force", builderNames=["testy"])]

        f = util.BuildFactory()
        f.addStep(SleepAndInterrupt())
        c['builders'] = [util.BuilderConfig(name="testy", workernames=["local1"], factory=f)]

        yield self.setup_master(c)

    @flaky(bugNumber=4404, onPlatform='win32')
    @defer.inlineCallbacks
    def test_interrupt(self):
        yield self.setup_config()
        build = yield self.doForceBuild(wantSteps=True)
        self.assertEqual(build['steps'][-1]['results'], CANCELLED)


class InterruptCommandPb(InterruptCommand):
    proto = "pb"


class InterruptCommandMsgPack(InterruptCommand):
    proto = "msgpack"



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import bz2
import zlib

from buildbot.db.compression.protocol import CompressObjInterface
from buildbot.db.compression.protocol import CompressorInterface


class GZipCompressor(CompressorInterface):
    name = "gz"

    COMPRESS_LEVEL = zlib.Z_BEST_COMPRESSION

    @staticmethod
    def dumps(data: bytes) -> bytes:
        return zlib.compress(data, level=GZipCompressor.COMPRESS_LEVEL)

    @staticmethod
    def read(data: bytes) -> bytes:
        return zlib.decompress(data)

    class CompressObj(CompressObjInterface):
        def __init__(self) -> None:
            self._create_compressobj()

        def _create_compressobj(self) -> None:
            self._compressobj = zlib.compressobj(level=GZipCompressor.COMPRESS_LEVEL)

        def compress(self, data: bytes) -> bytes:
            return self._compressobj.compress(data)

        def flush(self) -> bytes:
            try:
                return self._compressobj.flush(zlib.Z_FINISH)
            finally:
                # recreate compressobj so this instance can be re-used
                self._create_compressobj()


class BZipCompressor(CompressorInterface):
    name = "bz2"

    COMPRESS_LEVEL = 9

    @staticmethod
    def dumps(data: bytes) -> bytes:
        return bz2.compress(data, BZipCompressor.COMPRESS_LEVEL)

    @staticmethod
    def read(data: bytes) -> bytes:
        return bz2.decompress(data)

    class CompressObj(CompressObjInterface):
        def __init__(self) -> None:
            self._create_compressobj()

        def _create_compressobj(self) -> None:
            self._compressobj = bz2.BZ2Compressor(BZipCompressor.COMPRESS_LEVEL)

        def compress(self, data: bytes) -> bytes:
            return self._compressobj.compress(data)

        def flush(self) -> bytes:
            try:
                return self._compressobj.flush()
            finally:
                # recreate compressobj so this instance can be re-used
                self._create_compressobj()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any

from buildbot.data import base
from buildbot.data import types
from buildbot.util.twisted import async_to_deferred

if TYPE_CHECKING:
    from buildbot.db.codebase_commits import CodebaseCommitModel


def commit_db_to_data(model: CodebaseCommitModel) -> dict[str, Any]:
    return {
        "commitid": model.id,
        "codebaseid": model.codebaseid,
        'author': model.author,
        'committer': model.committer,
        'comments': model.comments,
        'when_timestamp': model.when_timestamp,
        'revision': model.revision,
        'parent_commitid': model.parent_commitid,
    }


commits_field_map = {
    'commitid': 'codebase_commits.id',
    'codebaseid': 'codebase_commits.codebaseid',
    'author': 'codebase_commits.author',
    'committer': 'codebase_commits.committer',
    'comments': 'codebase_commits.comments',
    'when_timestamp': 'codebase_commits.when_timestamp',
    'parent_commitid': 'codebase_commits.parent_commitid',
}


class CodebaseCommitEndpoint(base.Endpoint):
    kind = base.EndpointKind.SINGLE
    pathPatterns = [
        "/commits/n:commitid",
    ]

    @async_to_deferred
    async def get(self, result_spec: base.ResultSpec, kwargs: Any) -> dict[str, Any] | None:
        commit = await self.master.db.codebase_commits.get_commit(kwargs['commitid'])
        if not commit:
            return None
        return commit_db_to_data(commit)


class CodebaseCommitByRevisionEndpoint(base.Endpoint):
    kind = base.EndpointKind.SINGLE
    pathPatterns = [
        "/codebases/n:codebaseid/commits_by_revision/s:revision",
    ]

    @async_to_deferred
    async def get(self, result_spec: base.ResultSpec, kwargs: Any) -> dict[str, Any] | None:
        commit = await self.master.db.codebase_commits.get_commit_by_revision(
            codebaseid=kwargs['codebaseid'], revision=kwargs['revision']
        )
        if not commit:
            return None
        return commit_db_to_data(commit)


class CodebaseCommitsEndpoint(base.Endpoint):
    kind = base.EndpointKind.COLLECTION
    pathPatterns = [
        "/codebases/n:codebaseid/commits",
    ]

    @async_to_deferred
    async def get(self, result_spec: base.ResultSpec, kwargs: Any) -> list[dict[str, Any]]:
        result_spec.fieldMapping = commits_field_map
        commits = await self.master.db.codebase_commits.get_commits(
            codebaseid=kwargs['codebaseid'], result_spec=result_spec
        )
        return [commit_db_to_data(c) for c in commits]


class CodebaseCommit(base.ResourceType):
    name = "commit"
    plural = "commits"
    endpoints = [CodebaseCommitEndpoint, CodebaseCommitByRevisionEndpoint, CodebaseCommitsEndpoint]
    eventPathPatterns = [
        "/commits/:commitid",
        "/codebases/:codebaseid/commits/:commitid",
    ]

    class EntityType(types.Entity):
        commitid = types.Integer()
        codebaseid = types.Integer()
        author = types.String()
        committer = types.NoneOk(types.String())
        comments = types.String()
        when_timestamp = types.Integer()
        revision = types.String()
        parent_commitid = types.NoneOk(types.Integer())

    entityType = EntityType(name)

    @async_to_deferred
    async def generate_event(self, _id: int, event: str) -> None:
        commit = await self.master.data.get(('commits', str(_id)))
        self.produceEvent(commit, event)

    @base.updateMethod
    @async_to_deferred
    async def add_commit(
        self,
        *,
        codebaseid: int,
        author: str,
        committer: str | None = None,
        files: list[str] | None = None,
        comments: str,
        when_timestamp: int,
        revision: str,
        parent_commitid: int | None = None,
    ) -> None:
        commitid = await self.master.db.codebase_commits.add_commit(
            codebaseid=codebaseid,
            author=author,
            committer=committer,
            files=files,
            comments=comments,
            when_timestamp=when_timestamp,
            revision=revision,
            parent_commitid=parent_commitid,
        )
        await self.generate_event(commitid, 'new')
        return commitid


class CodebaseCommitsGraphEndpoint(base.Endpoint):
    kind = base.EndpointKind.SINGLE
    pathPatterns = [
        "/codebases/n:codebaseid/commit_range/n:commitid1/n:commitid2/commits_common_parent",
    ]

    @async_to_deferred
    async def get(self, resultSpec: base.ResultSpec, kwargs: Any) -> dict[str, Any] | None:
        r = await self.master.db.codebase_commits.get_first_common_commit_with_ranges(
            int(kwargs.get('commitid1')), int(kwargs.get('commitid2'))
        )

        if r is None:
            return None
        return {
            'common': r.common_commit_id,
            'to1': r.to1_commit_ids,
            'to2': r.to2_commit_ids,
        }


class CodebaseCommitsGraph(base.ResourceType):
    name = "commit_graph"
    plural = "commit_graphs"
    endpoints = [CodebaseCommitsGraphEndpoint]

    class EntityType(types.Entity):
        common = types.Integer()
        to1 = types.List(of=types.Integer())
        to2 = types.List(of=types.Integer())

    entityType = EntityType(name)
